# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from oslo_log import log as logging

from cinder import exception
from cinder.i18n import _
from cinder.volume.drivers.dell_emc.powermax import masking
from cinder.volume.drivers.dell_emc.powermax import provision
from cinder.volume.drivers.dell_emc.powermax import utils

LOG = logging.getLogger(__name__)


class PowerMaxMigrate(object):
    """Upgrade class for Rest based PowerMax volume drivers.

    This upgrade class is for Dell EMC PowerMax volume drivers
    based on UniSphere Rest API.
    It supports VMAX 3 and VMAX All Flash and PowerMax arrays.

    """
    def __init__(self, prtcl, rest):
        self.rest = rest
        self.utils = utils.PowerMaxUtils()
        self.masking = masking.PowerMaxMasking(prtcl, self.rest)
        self.provision = provision.PowerMaxProvision(self.rest)

    def do_migrate_if_candidate(
            self, array, srp, device_id, volume, connector):
        """Check and migrate if the volume is a candidate

        If the volume is in the legacy (SMIS) masking view structure
        move it to staging storage group within a staging masking view.

        :param array: array serial number
        :param srp: the SRP
        :param device_id: the volume device id
        :param volume: the volume object
        :param connector: the connector object
        """
        mv_detail_list = list()

        masking_view_list, storage_group_list = (
            self._get_mvs_and_sgs_from_volume(
                array, device_id))

        for masking_view in masking_view_list:
            masking_view_dict = self.get_masking_view_component_dict(
                masking_view, srp)
            if masking_view_dict:
                mv_detail_list.append(masking_view_dict)

        if not mv_detail_list:
            return False

        if len(storage_group_list) != 1:
            LOG.warning("MIGRATE - The volume %(dev_id)s is not in one "
                        "storage group as is expected for migration. "
                        "The volume is in storage groups %(sg_list)s."
                        "Migration will not proceed.",
                        {'dev_id': device_id,
                         'sg_list': storage_group_list})
            return False
        else:
            source_storage_group_name = storage_group_list[0]

        # Get the host that OpenStack has volume exposed to (it should only
        # be one host).
        os_host_list = self.get_volume_host_list(volume, connector)
        if len(os_host_list) != 1:
            LOG.warning("MIGRATE - OpenStack has recorded that "
                        "%(dev_id)s is attached to hosts %(os_hosts)s "
                        "and not 1 host as is expected. "
                        "Migration will not proceed.",
                        {'dev_id': device_id,
                         'os_hosts': os_host_list})
            return False
        else:
            os_host_name = os_host_list[0]
        LOG.info("MIGRATE - Volume %(dev_id)s is a candidate for "
                 "migration. The OpenStack host is %(os_host_name)s."
                 "The volume is in storage group %(sg_name)s.",
                 {'dev_id': device_id,
                  'os_host_name': os_host_name,
                  'sg_name': source_storage_group_name})
        return self._perform_migration(
            array, device_id, mv_detail_list, source_storage_group_name,
            os_host_name)

    def _perform_migration(
            self, array, device_id, mv_detail_list, source_storage_group_name,
            os_host_name):
        """Perform steps so we can get the volume in a correct state.

        :param array: the storage array
        :param device_id: the device_id
        :param mv_detail_list: the masking view list
        :param source_storage_group_name: the source storage group
        :param os_host_name: the host the volume is exposed to
        :returns: boolean
        """
        extra_specs = {utils.INTERVAL: 3, utils.RETRIES: 200}
        stg_sg_name = self._create_stg_storage_group_with_vol(
            array, os_host_name, extra_specs)
        if not stg_sg_name:
            # Throw an exception here
            exception_message = _("MIGRATE - Unable to create staging "
                                  "storage group.")
            LOG.error(exception_message)
            raise exception.VolumeBackendAPIException(
                message=exception_message)
        LOG.info("MIGRATE - Staging storage group %(stg_sg_name)s has "
                 "been successfully created.", {'stg_sg_name': stg_sg_name})

        new_stg_mvs = self._create_stg_masking_views(
            array, mv_detail_list, stg_sg_name, extra_specs)
        LOG.info("MIGRATE - Staging masking views %(new_stg_mvs)s have "
                 "been successfully created.", {'new_stg_mvs': new_stg_mvs})

        if not new_stg_mvs:
            exception_message = _("MIGRATE - Unable to create staging "
                                  "masking views.")
            LOG.error(exception_message)
            raise exception.VolumeBackendAPIException(
                message=exception_message)

        # Move volume from old storage group to new staging storage group
        self.move_volume_from_legacy_to_staging(
            array, device_id, source_storage_group_name,
            stg_sg_name, extra_specs)

        LOG.info("MIGRATE - Device id %(device_id)s has been successfully "
                 "moved from %(src_sg)s to %(tgt_sg)s.",
                 {'device_id': device_id,
                  'src_sg': source_storage_group_name,
                  'tgt_sg': stg_sg_name})

        new_masking_view_list, new_storage_group_list = (
            self._get_mvs_and_sgs_from_volume(
                array, device_id))

        if len(new_storage_group_list) != 1:
            exception_message = (_(
                "MIGRATE - The current storage group list has %(list_len)d "
                "members. The list is %(sg_list)s. Will not proceed with "
                "cleanup. Please contact customer representative.") % {
                'list_len': len(new_storage_group_list),
                'sg_list': new_storage_group_list})
            LOG.error(exception_message)
            raise exception.VolumeBackendAPIException(
                message=exception_message)
        else:
            current_storage_group_name = new_storage_group_list[0]
            if current_storage_group_name.lower() != stg_sg_name.lower():
                exception_message = (_(
                    "MIGRATE - The current storage group %(sg_1)s "
                    "does not match %(sg_2)s. Will not proceed with "
                    "cleanup. Please contact customer representative.") % {
                    'sg_1': current_storage_group_name,
                    'sg_2': stg_sg_name})
                LOG.error(exception_message)
                raise exception.VolumeBackendAPIException(
                    message=exception_message)

        if not self._delete_staging_masking_views(
                array, new_masking_view_list, os_host_name):
            exception_message = _("MIGRATE - Unable to delete staging masking "
                                  "views. Please contact customer "
                                  "representative.")
            LOG.error(exception_message)
            raise exception.VolumeBackendAPIException(
                message=exception_message)

        final_masking_view_list, final_storage_group_list = (
            self._get_mvs_and_sgs_from_volume(
                array, device_id))
        if len(final_masking_view_list) != 1:
            exception_message = (_(
                "MIGRATE - The final masking view list has %(list_len)d "
                "entries and not 1 entry as is expected.  The list is "
                "%(mv_list)s. Please contact customer representative.") % {
                'list_len': len(final_masking_view_list),
                'sg_list': final_masking_view_list})
            LOG.error(exception_message)
            raise exception.VolumeBackendAPIException(
                message=exception_message)

        return True

    def move_volume_from_legacy_to_staging(
            self, array, device_id, source_storage_group_name,
            stg_sg_name, extra_specs):
        """Move the volume from legacy SG to staging SG

        :param array: array serial number
        :param device_id: the device id of the volume
        :param source_storage_group_name: the source storage group
        :param stg_sg_name: the target staging storage group
        :param extra_specs: the extra specs
        """
        num_vol_in_sg = self.rest.get_num_vols_in_sg(
            array, source_storage_group_name)
        if num_vol_in_sg == 1:
            # Can't move last volume and leave masking view empty
            # so creating a holder volume
            temp_vol_size = '1'
            hold_vol_name = 'hold-' + str(uuid.uuid1())
            self.provision.create_volume_from_sg(
                array, hold_vol_name, source_storage_group_name,
                temp_vol_size, extra_specs)
            LOG.info("MIGRATE - Volume %(vol)s has been created because "
                     "there was only one volume remaining in storage group "
                     "%(src_sg)s and we are attempting a move it to staging "
                     "storage group %(tgt_sg)s.",
                     {'vol': hold_vol_name,
                      'src_sg': source_storage_group_name,
                      'tgt_sg': stg_sg_name})

        self.rest.move_volume_between_storage_groups(
            array, device_id, source_storage_group_name,
            stg_sg_name, extra_specs)

    def _delete_staging_masking_views(
            self, array, masking_view_list, os_host_name):
        """Delete the staging masking views

        Delete the staging masking views except the masking view
        exposed to the OpenStack compute

        :param array: array serial number
        :param masking_view_list: masking view namelist
        :param os_host_name: the host the volume is exposed to in OpenStack
        :returns: boolean
        """
        delete_mv_list = list()
        safe_to_delete = False
        for masking_view_name in masking_view_list:
            if os_host_name in masking_view_name:
                safe_to_delete = True
            else:
                delete_mv_list.append(masking_view_name)
        if safe_to_delete:
            for delete_mv in delete_mv_list:
                self.rest.delete_masking_view(array, delete_mv)
                LOG.info("MIGRATE - Masking view %(delete_mv)s has been "
                         "successfully deleted.",
                         {'delete_mv': delete_mv})
        return safe_to_delete

    def _create_stg_masking_views(
            self, array, mv_detail_list, stg_sg_name, extra_specs):
        """Create a staging masking views

        :param array: array serial number
        :param mv_detail_list: masking view detail list
        :param stg_sg_name: staging storage group name
        :param extra_specs: the extra specs
        :returns: masking view list
        """
        new_masking_view_list = list()
        for mv_detail in mv_detail_list:
            host_name = mv_detail.get('host')
            masking_view_name = mv_detail.get('mv_name')
            masking_view_components = self.rest.get_masking_view(
                array, masking_view_name)
            # Create a staging masking view
            random_uuid = uuid.uuid1()
            staging_mv_name = 'STG-' + host_name + '-' + str(
                random_uuid) + '-MV'
            if masking_view_components:
                self.rest.create_masking_view(
                    array, staging_mv_name, stg_sg_name,
                    masking_view_components.get('portGroupId'),
                    masking_view_components.get('hostId'), extra_specs)
                masking_view_dict = self.rest.get_masking_view(
                    array, staging_mv_name)
                if masking_view_dict:
                    new_masking_view_list.append(staging_mv_name)
                else:
                    LOG.warning("Failed to create staging masking view "
                                "%(mv_name)s. Migration cannot proceed.",
                                {'mv_name': masking_view_name})
                    return None
        return new_masking_view_list

    def _create_stg_storage_group_with_vol(self, array, os_host_name,
                                           extra_specs):
        """Create a staging storage group and add volume

        :param array: array serial number
        :param os_host_name: the openstack host name
        :param extra_specs: the extra specs
        :returns: storage group name
        """
        random_uuid = uuid.uuid1()
        # Create a staging SG
        stg_sg_name = 'STG-' + os_host_name + '-' + (
            str(random_uuid) + '-SG')
        temp_vol_name = 'tempvol-' + str(random_uuid)
        temp_vol_size = '1'

        _stg_storage_group = self.provision.create_storage_group(
            array, stg_sg_name,
            None, None, None, extra_specs)
        if _stg_storage_group:
            self.provision.create_volume_from_sg(
                array, temp_vol_name, stg_sg_name,
                temp_vol_size, extra_specs)
            return stg_sg_name
        else:
            return None

    def _get_mvs_and_sgs_from_volume(self, array, device_id):
        """Given a device Id get its storage groups and masking views.

        :param array: array serial number
        :param device_id: the volume device id
        :returns: masking view list, storage group list
        """
        final_masking_view_list = []
        storage_group_list = self.rest.get_storage_groups_from_volume(
            array, device_id)
        for sg in storage_group_list:
            masking_view_list = self.rest.get_masking_views_from_storage_group(
                array, sg)
            final_masking_view_list.extend(masking_view_list)
        return final_masking_view_list, storage_group_list

    def get_masking_view_component_dict(
            self, masking_view_name, srp):
        """Get components from input string.

        :param masking_view_name: the masking view name -- str
        :param srp: the srp -- str
        :returns: object components -- dict
        """
        regex_str_share = (
            r'^(?P<prefix>OS)-(?P<host>.+?)((?P<srp>' + srp + r')-'
            r'(?P<slo>.+?)-(?P<workload>.+?)|(?P<no_slo>No_SLO))'
            r'((?P<protocol>-I|-F)|)'
            r'(?P<CD>-CD|)(?P<RE>-RE|)'
            r'(?P<uuid>-[0-9A-Fa-f]{8}|)'
            r'-(?P<postfix>MV)$')

        object_dict = self.utils.get_object_components_and_correct_host(
            regex_str_share, masking_view_name)

        if object_dict:
            object_dict['mv_name'] = masking_view_name
        return object_dict

    def get_volume_host_list(self, volume, connector):
        """Get host list attachments from connector object

        :param volume: the volume object
        :param connector: the connector object
        :returns os_host_list
        """
        os_host_list = list()
        if connector is not None:
            attachment_list = volume.volume_attachment
            LOG.debug("Volume attachment list: %(atl)s. "
                      "Attachment type: %(at)s",
                      {'atl': attachment_list, 'at': type(attachment_list)})
            try:
                att_list = attachment_list.objects
            except AttributeError:
                att_list = attachment_list
            if att_list is not None:
                host_list = [att.connector['host'] for att in att_list if
                             att is not None and att.connector is not None]
        for host_name in host_list:
            os_host_list.append(self.utils.get_host_short_name(host_name))
        return os_host_list

    def cleanup_staging_objects(
            self, array, storage_group_names, extra_specs):
        """Delete the staging masking views and storage groups

        :param array: the array serial number
        :param storage_group_names: a list of storage group names
        :param extra_specs: the extra specs
        """
        def _do_cleanup(sg_name, device_id):
            masking_view_list = (
                self.rest.get_masking_views_from_storage_group(
                    array, sg_name))
            for masking_view in masking_view_list:
                if 'STG-' in masking_view:
                    self.rest.delete_masking_view(array, masking_view)
                    self.rest.remove_vol_from_sg(
                        array, sg_name, device_id,
                        extra_specs)
                    self.rest.delete_volume(array, device_id)
                    self.rest.delete_storage_group(array, sg_name)

        for storage_group_name in storage_group_names:
            if 'STG-' in storage_group_name:
                volume_list = self.rest.get_volumes_in_storage_group(
                    array, storage_group_name)
                if len(volume_list) == 1:
                    try:
                        _do_cleanup(storage_group_name, volume_list[0])
                    except Exception:
                        LOG.warning("MIGRATE - An attempt was made to "
                                    "cleanup after a legacy live migration, "
                                    "but it failed. You may choose to "
                                    "cleanup manually.")



# Copyright (c) 2016 Clinton Knight
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy

import ddt
import mock
import six

from cinder import exception
from cinder import test
from cinder.tests.unit.volume.drivers.netapp.dataontap.client import (
    fakes as fake_client)
import cinder.tests.unit.volume.drivers.netapp.dataontap.utils.fakes as fake
import cinder.tests.unit.volume.drivers.netapp.fakes as na_fakes
from cinder.volume.drivers.netapp.dataontap.utils import capabilities


@ddt.ddt
class CapabilitiesLibraryTestCase(test.TestCase):

    def setUp(self):
        super(CapabilitiesLibraryTestCase, self).setUp()

        self.zapi_client = mock.Mock()
        self.configuration = self.get_config_cmode()
        self.ssc_library = capabilities.CapabilitiesLibrary(
            'iSCSI', fake.SSC_VSERVER, self.zapi_client, self.configuration)
        self.ssc_library.ssc = fake.SSC

    def get_config_cmode(self):
        config = na_fakes.create_configuration_cmode()
        config.volume_backend_name = 'fake_backend'
        return config

    def test_check_api_permissions(self):

        mock_log = self.mock_object(capabilities.LOG, 'warning')

        self.ssc_library.check_api_permissions()

        self.zapi_client.check_cluster_api.assert_has_calls(
            [mock.call(*key) for key in capabilities.SSC_API_MAP.keys()])
        self.assertEqual(0, mock_log.call_count)

    def test_check_api_permissions_failed_ssc_apis(self):

        def check_cluster_api(object_name, operation_name, api):
            if api != 'volume-get-iter':
                return False
            return True

        self.zapi_client.check_cluster_api.side_effect = check_cluster_api
        mock_log = self.mock_object(capabilities.LOG, 'warning')

        self.ssc_library.check_api_permissions()

        self.assertEqual(1, mock_log.call_count)

    def test_check_api_permissions_failed_volume_api(self):

        def check_cluster_api(object_name, operation_name, api):
            if api == 'volume-get-iter':
                return False
            return True

        self.zapi_client.check_cluster_api.side_effect = check_cluster_api
        mock_log = self.mock_object(capabilities.LOG, 'warning')

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.ssc_library.check_api_permissions)

        self.assertEqual(0, mock_log.call_count)

    def test_get_ssc(self):

        result = self.ssc_library.get_ssc()

        self.assertEqual(fake.SSC, result)
        self.assertIsNot(fake.SSC, result)

    def test_get_ssc_flexvol_names(self):

        result = self.ssc_library.get_ssc_flexvol_names()

        self.assertItemsEqual(fake.SSC_VOLUMES, result)

    def test_get_ssc_for_flexvol(self):

        result = self.ssc_library.get_ssc_for_flexvol(fake.SSC_VOLUMES[0])

        self.assertEqual(fake.SSC.get(fake.SSC_VOLUMES[0]), result)
        self.assertIsNot(fake.SSC.get(fake.SSC_VOLUMES[0]), result)

    def test_get_ssc_for_flexvol_not_found(self):

        result = self.ssc_library.get_ssc_for_flexvol('invalid')

        self.assertEqual({}, result)

    def test_get_ssc_aggregates(self):

        result = self.ssc_library.get_ssc_aggregates()

        six.assertCountEqual(self, list(fake.SSC_AGGREGATES), result)

    def test_update_ssc(self):

        mock_get_ssc_flexvol_info = self.mock_object(
            self.ssc_library, '_get_ssc_flexvol_info',
            side_effect=[fake.SSC_FLEXVOL_INFO['volume1'],
                         fake.SSC_FLEXVOL_INFO['volume2']])
        mock_get_ssc_dedupe_info = self.mock_object(
            self.ssc_library, '_get_ssc_dedupe_info',
            side_effect=[fake.SSC_DEDUPE_INFO['volume1'],
                         fake.SSC_DEDUPE_INFO['volume2']])
        mock_get_ssc_mirror_info = self.mock_object(
            self.ssc_library, '_get_ssc_mirror_info',
            side_effect=[fake.SSC_MIRROR_INFO['volume1'],
                         fake.SSC_MIRROR_INFO['volume2']])
        mock_get_ssc_aggregate_info = self.mock_object(
            self.ssc_library, '_get_ssc_aggregate_info',
            side_effect=[fake.SSC_AGGREGATE_INFO['volume1'],
                         fake.SSC_AGGREGATE_INFO['volume2']])
        mock_get_ssc_encryption_info = self.mock_object(
            self.ssc_library, '_get_ssc_encryption_info',
            side_effect=[fake.SSC_ENCRYPTION_INFO['volume1'],
                         fake.SSC_ENCRYPTION_INFO['volume2']])
        ordered_ssc = collections.OrderedDict()
        ordered_ssc['volume1'] = fake.SSC_VOLUME_MAP['volume1']
        ordered_ssc['volume2'] = fake.SSC_VOLUME_MAP['volume2']

        result = self.ssc_library.update_ssc(ordered_ssc)

        self.assertIsNone(result)
        self.assertEqual(fake.SSC, self.ssc_library.ssc)
        mock_get_ssc_flexvol_info.assert_has_calls([
            mock.call('volume1'), mock.call('volume2')])
        mock_get_ssc_dedupe_info.assert_has_calls([
            mock.call('volume1'), mock.call('volume2')])
        mock_get_ssc_mirror_info.assert_has_calls([
            mock.call('volume1'), mock.call('volume2')])
        mock_get_ssc_aggregate_info.assert_has_calls([
            mock.call('aggr1'), mock.call('aggr2')])
        mock_get_ssc_encryption_info.assert_has_calls([
            mock.call('volume1'), mock.call('volume2')])

    def test__update_for_failover(self):
        self.mock_object(self.ssc_library, 'update_ssc')
        flexvol_map = {'volume1': fake.SSC_VOLUME_MAP['volume1']}
        mock_client = mock.Mock(name='FAKE_ZAPI_CLIENT')

        self.ssc_library._update_for_failover(mock_client, flexvol_map)

        self.assertEqual(mock_client, self.ssc_library.zapi_client)
        self.ssc_library.update_ssc.assert_called_once_with(flexvol_map)

    @ddt.data({'lun_space_guarantee': True},
              {'lun_space_guarantee': False})
    @ddt.unpack
    def test_get_ssc_flexvol_info_thin_block(self, lun_space_guarantee):

        self.ssc_library.configuration.netapp_lun_space_reservation = \
            'enabled' if lun_space_guarantee else 'disabled'
        self.mock_object(self.ssc_library.zapi_client,
                         'get_flexvol',
                         return_value=fake_client.VOLUME_INFO_SSC)

        result = self.ssc_library._get_ssc_flexvol_info(
            fake_client.VOLUME_NAMES[0])

        expected = {
            'netapp_thin_provisioned': 'true',
            'thick_provisioning_support': False,
            'thin_provisioning_support': True,
            'netapp_aggregate': 'fake_aggr1',
        }
        self.assertEqual(expected, result)
        self.zapi_client.get_flexvol.assert_called_once_with(
            flexvol_name=fake_client.VOLUME_NAMES[0])

    @ddt.data({'vol_space_guarantee': 'file', 'lun_space_guarantee': True},
              {'vol_space_guarantee': 'volume', 'lun_space_guarantee': True})
    @ddt.unpack
    def test_get_ssc_flexvol_info_thick_block(self, vol_space_guarantee,
                                              lun_space_guarantee):

        self.ssc_library.configuration.netapp_lun_space_reservation = \
            'enabled' if lun_space_guarantee else 'disabled'
        fake_volume_info_ssc = copy.deepcopy(fake_client.VOLUME_INFO_SSC)
        fake_volume_info_ssc['space-guarantee'] = vol_space_guarantee
        self.mock_object(self.ssc_library.zapi_client,
                         'get_flexvol',
                         return_value=fake_volume_info_ssc)

        result = self.ssc_library._get_ssc_flexvol_info(
            fake_client.VOLUME_NAMES[0])

        expected = {
            'netapp_thin_provisioned': 'false',
            'thick_provisioning_support': lun_space_guarantee,
            'thin_provisioning_support': not lun_space_guarantee,
            'netapp_aggregate': 'fake_aggr1',
        }
        self.assertEqual(expected, result)
        self.zapi_client.get_flexvol.assert_called_once_with(
            flexvol_name=fake_client.VOLUME_NAMES[0])

    @ddt.data({'nfs_sparsed_volumes': True},
              {'nfs_sparsed_volumes': False})
    @ddt.unpack
    def test_get_ssc_flexvol_info_thin_file(self, nfs_sparsed_volumes):

        self.ssc_library.protocol = 'nfs'
        self.ssc_library.configuration.nfs_sparsed_volumes = \
            nfs_sparsed_volumes
        self.mock_object(self.ssc_library.zapi_client,
                         'get_flexvol',
                         return_value=fake_client.VOLUME_INFO_SSC)

        result = self.ssc_library._get_ssc_flexvol_info(
            fake_client.VOLUME_NAMES[0])

        expected = {
            'netapp_thin_provisioned': 'true',
            'thick_provisioning_support': False,
            'thin_provisioning_support': True,
            'netapp_aggregate': 'fake_aggr1',
        }
        self.assertEqual(expected, result)
        self.zapi_client.get_flexvol.assert_called_once_with(
            flexvol_name=fake_client.VOLUME_NAMES[0])

    @ddt.data({'vol_space_guarantee': 'file', 'nfs_sparsed_volumes': True},
              {'vol_space_guarantee': 'volume', 'nfs_sparsed_volumes': False})
    @ddt.unpack
    def test_get_ssc_flexvol_info_thick_file(self, vol_space_guarantee,
                                             nfs_sparsed_volumes):

        self.ssc_library.protocol = 'nfs'
        self.ssc_library.configuration.nfs_sparsed_volumes = \
            nfs_sparsed_volumes
        fake_volume_info_ssc = copy.deepcopy(fake_client.VOLUME_INFO_SSC)
        fake_volume_info_ssc['space-guarantee'] = vol_space_guarantee
        self.mock_object(self.ssc_library.zapi_client,
                         'get_flexvol',
                         return_value=fake_volume_info_ssc)

        result = self.ssc_library._get_ssc_flexvol_info(
            fake_client.VOLUME_NAMES[0])

        expected = {
            'netapp_thin_provisioned': 'false',
            'thick_provisioning_support': not nfs_sparsed_volumes,
            'thin_provisioning_support': nfs_sparsed_volumes,
            'netapp_aggregate': 'fake_aggr1',
        }
        self.assertEqual(expected, result)
        self.zapi_client.get_flexvol.assert_called_once_with(
            flexvol_name=fake_client.VOLUME_NAMES[0])

    @ddt.data([], ['netapp_dedup'], ['netapp_compression'])
    def test_get_ssc_dedupe_info(self, invalid_extra_specs):
        self.ssc_library.invalid_extra_specs = invalid_extra_specs
        self.mock_object(
            self.ssc_library.zapi_client, 'get_flexvol_dedupe_info',
            return_value=fake_client.VOLUME_DEDUPE_INFO_SSC)

        result = self.ssc_library._get_ssc_dedupe_info(
            fake_client.VOLUME_NAMES[0])

        if invalid_extra_specs:
            expected = {
                'netapp_dedup': 'false',
                'netapp_compression': 'false',
            }
            self.zapi_client.get_flexvol_dedupe_info.assert_not_called()
        else:
            expected = {
                'netapp_dedup': 'true',
                'netapp_compression': 'false',
            }
            self.zapi_client.get_flexvol_dedupe_info.assert_called_once_with(
                fake_client.VOLUME_NAMES[0])
        self.assertEqual(expected, result)

    def test_get_ssc_encryption_info(self):

        self.mock_object(
            self.ssc_library.zapi_client, 'is_flexvol_encrypted',
            return_value=True)

        result = self.ssc_library._get_ssc_encryption_info(
            fake_client.VOLUME_NAMES[0])

        expected = {
            'netapp_flexvol_encryption': 'true',
        }
        self.assertEqual(expected, result)
        self.zapi_client.is_flexvol_encrypted.assert_called_once_with(
            fake_client.VOLUME_NAMES[0], fake_client.VOLUME_VSERVER_NAME)

    @ddt.data(True, False)
    def test_get_ssc_mirror_info(self, mirrored):

        self.mock_object(
            self.ssc_library.zapi_client, 'is_flexvol_mirrored',
            return_value=mirrored)

        result = self.ssc_library._get_ssc_mirror_info(
            fake_client.VOLUME_NAMES[0])

        expected = {'netapp_mirrored': 'true' if mirrored else 'false'}
        self.assertEqual(expected, result)
        self.zapi_client.is_flexvol_mirrored.assert_called_once_with(
            fake_client.VOLUME_NAMES[0], fake.SSC_VSERVER)

    @ddt.data([], ['netapp_raid_type'])
    def test_get_ssc_aggregate_info(self, invalid_extra_specs):
        self.ssc_library.invalid_extra_specs = invalid_extra_specs
        self.mock_object(
            self.ssc_library.zapi_client, 'get_aggregate',
            return_value=fake_client.AGGR_INFO_SSC)
        self.mock_object(
            self.ssc_library.zapi_client, 'get_aggregate_disk_types',
            return_value=fake_client.AGGREGATE_DISK_TYPES)

        result = self.ssc_library._get_ssc_aggregate_info(
            fake_client.VOLUME_AGGREGATE_NAME)

        if invalid_extra_specs:
            expected = {
                'netapp_disk_type': None,
                'netapp_raid_type': None,
                'netapp_hybrid_aggregate': None,
            }
            self.zapi_client.get_aggregate.assert_not_called()
            self.zapi_client.get_aggregate_disk_types.assert_not_called()
        else:
            expected = {
                'netapp_disk_type': fake_client.AGGREGATE_DISK_TYPES,
                'netapp_raid_type': fake_client.AGGREGATE_RAID_TYPE,
                'netapp_hybrid_aggregate': 'true',
            }
            self.zapi_client.get_aggregate.assert_called_once_with(
                fake_client.VOLUME_AGGREGATE_NAME)
            self.zapi_client.get_aggregate_disk_types.assert_called_once_with(
                fake_client.VOLUME_AGGREGATE_NAME)

        self.assertEqual(expected, result)

    def test_get_ssc_aggregate_info_not_found(self):
        self.ssc_library.invalid_extra_specs = ['netapp_raid_type']
        self.mock_object(
            self.ssc_library.zapi_client, 'get_aggregate', return_value={})
        self.mock_object(
            self.ssc_library.zapi_client, 'get_aggregate_disk_types',
            return_value=None)

        result = self.ssc_library._get_ssc_aggregate_info(
            fake_client.VOLUME_AGGREGATE_NAME)

        expected = {
            'netapp_disk_type': None,
            'netapp_raid_type': None,
            'netapp_hybrid_aggregate': None,
        }
        self.assertEqual(expected, result)

    def test_get_matching_flexvols_for_extra_specs(self):

        specs = {
            'thick_provisioning_support': '<is> False',
            'netapp_compression': 'true',
            'netapp_dedup': 'true',
            'netapp_mirrored': 'true',
            'netapp_raid_type': 'raid_dp',
            'netapp_disk_type': 'FCAL',
            'non_ssc_key': 'fake_value',
        }

        result = self.ssc_library.get_matching_flexvols_for_extra_specs(specs)

        self.assertEqual(['volume2'], result)

    @ddt.data(
        {
            'flexvol_info': {
                'netapp_dedup': 'true',
            },
            'extra_specs': {
                'netapp_dedup': 'true',
                'non_ssc_key': 'fake_value',
            }
        },
        {
            'flexvol_info': fake.SSC['volume1'],
            'extra_specs': {
                'netapp_disk_type': 'SSD',
                'pool_name': 'volume1',
            }
        },
        {
            'flexvol_info': fake.SSC['volume2'],
            'extra_specs': {
                'netapp_disk_type': 'SSD',
                'netapp_hybrid_aggregate': 'true',
            }
        }
    )
    @ddt.unpack
    def test_flexvol_matches_extra_specs(self, flexvol_info, extra_specs):

        result = self.ssc_library._flexvol_matches_extra_specs(flexvol_info,
                                                               extra_specs)

        self.assertTrue(result)

    @ddt.data(
        {
            'flexvol_info': {
                'netapp_dedup': 'true',
            },
            'extra_specs': {
                'netapp_dedup': 'false',
                'non_ssc_key': 'fake_value',
            }
        },
        {
            'flexvol_info': fake.SSC['volume2'],
            'extra_specs': {
                'netapp_disk_type': 'SSD',
                'pool_name': 'volume1',
            }
        },
        {
            'flexvol_info': fake.SSC['volume2'],
            'extra_specs': {
                'netapp_disk_type': 'SATA',
            }
        }
    )
    @ddt.unpack
    def test_flexvol_matches_extra_specs_no_match(self, flexvol_info,
                                                  extra_specs):

        result = self.ssc_library._flexvol_matches_extra_specs(flexvol_info,
                                                               extra_specs)

        self.assertFalse(result)

    @ddt.data(('SSD', 'SSD'), ('SSD', ['SSD', 'FCAL']))
    @ddt.unpack
    def test_extra_spec_matches(self, extra_spec_value, ssc_flexvol_value):

        result = self.ssc_library._extra_spec_matches(extra_spec_value,
                                                      ssc_flexvol_value)

        self.assertTrue(result)

    @ddt.data(('SSD', 'FCAL'), ('SSD', ['FCAL']))
    @ddt.unpack
    def test_extra_spec_matches_no_match(self, extra_spec_value,
                                         ssc_flexvol_value):

        result = self.ssc_library._extra_spec_matches(extra_spec_value,
                                                      ssc_flexvol_value)

        self.assertFalse(result)

    def test_modify_extra_specs_for_comparison(self):

        specs = {
            'thick_provisioning_support': '<is> False',
            'thin_provisioning_support': '<is>  true',
            'netapp_compression': 'true',
        }

        result = self.ssc_library._modify_extra_specs_for_comparison(specs)

        expected = {
            'thick_provisioning_support': False,
            'thin_provisioning_support': True,
            'netapp_compression': 'true',
        }
        self.assertEqual(expected, result)

    @ddt.data([], ['netapp_dedup'], ['netapp_compression'])
    def test_cluster_user_supported(self, invalid_extra_specs):
        self.ssc_library.invalid_extra_specs = invalid_extra_specs

        if invalid_extra_specs:
            self.assertFalse(self.ssc_library.cluster_user_supported())
        else:
            self.assertTrue(self.ssc_library.cluster_user_supported())



# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_i18n import fixture as i18n_fixture
from oslo_serialization import jsonutils
from six.moves import http_client
import webob.dec

from cinder.api.middleware import fault
from cinder.api.openstack import wsgi
from cinder import test


class TestFaults(test.TestCase):
    """Tests covering `cinder.api.openstack.faults:Fault` class."""

    def setUp(self):
        super(TestFaults, self).setUp()
        self.useFixture(i18n_fixture.ToggleLazy(True))

    def test_400_fault_json(self):
        """Test fault serialized to JSON via file-extension and/or header."""
        requests = [
            webob.Request.blank('/.json'),
            webob.Request.blank('/', headers={"Accept": "application/json"}),
        ]

        for request in requests:
            fault = wsgi.Fault(webob.exc.HTTPBadRequest(explanation='scram'))
            response = request.get_response(fault)

            expected = {
                "badRequest": {
                    "message": "scram",
                    "code": http_client.BAD_REQUEST,
                },
            }
            actual = jsonutils.loads(response.body)

            self.assertEqual("application/json", response.content_type)
            self.assertEqual(expected, actual)

    def test_413_fault_json(self):
        """Test fault serialized to JSON via file-extension and/or header."""
        requests = [
            webob.Request.blank('/.json'),
            webob.Request.blank('/', headers={"Accept": "application/json"}),
        ]

        for request in requests:
            exc = webob.exc.HTTPRequestEntityTooLarge
            fault = wsgi.Fault(exc(explanation='sorry',
                                   headers={'Retry-After': '4'}))
            response = request.get_response(fault)

            expected = {
                "overLimit": {
                    "message": "sorry",
                    "code": http_client.REQUEST_ENTITY_TOO_LARGE,
                    "retryAfter": "4",
                },
            }
            actual = jsonutils.loads(response.body)

            self.assertEqual("application/json", response.content_type)
            self.assertEqual(expected, actual)

    def test_fault_has_status_int(self):
        """Ensure the status_int is set correctly on faults."""
        fault = wsgi.Fault(webob.exc.HTTPBadRequest(explanation='what?'))
        self.assertEqual(http_client.BAD_REQUEST, fault.status_int)


class ExceptionTest(test.TestCase):

    def _wsgi_app(self, inner_app):
        return fault.FaultWrapper(inner_app)

    def test_unicode_decode_error(self):
        @webob.dec.wsgify
        def unicode_error(req):
            raise UnicodeDecodeError("ascii", b"", 0, 1, "bad")

        api = self._wsgi_app(unicode_error)
        resp = webob.Request.blank('/').get_response(api)
        self.assertEqual(400, resp.status_int)



# Copyright (c) 2018 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from oslo_log import log as logging
import requests
import six

from cinder import exception
from cinder.i18n import _
from cinder.volume.drivers.fusionstorage import constants

LOG = logging.getLogger(__name__)


class RestCommon(object):
    def __init__(self, fs_address, fs_user, fs_password):
        self.address = fs_address
        self.user = fs_user
        self.password = fs_password

        self.session = None
        self.token = None
        self.version = None

        self.init_http_head()

        LOG.warning("Suppressing requests library SSL Warnings")
        requests.packages.urllib3.disable_warnings(
            requests.packages.urllib3.exceptions.InsecureRequestWarning)
        requests.packages.urllib3.disable_warnings(
            requests.packages.urllib3.exceptions.InsecurePlatformWarning)

    def init_http_head(self):
        self.session = requests.Session()
        self.session.headers.update({
            "Content-Type": "application/json;charset=UTF-8",
        })
        self.session.verify = False

    def call(self, url, method, data=None,
             call_timeout=constants.DEFAULT_TIMEOUT,
             get_version=False, filter_flag=False, json_flag=False):
        kwargs = {'timeout': call_timeout}
        if data:
            kwargs['data'] = json.dumps(data)

        if not get_version:
            call_url = self.address + constants.BASIC_URI + self.version + url
        else:
            call_url = self.address + constants.BASIC_URI + url

        func = getattr(self.session, method.lower())

        try:
            result = func(call_url, **kwargs)
        except Exception as err:
            LOG.error('Bad response from server: %(url)s. '
                      'Error: %(err)s'), {'url': url, 'err': err}
            return {"error": {
                "code": constants.CONNECT_ERROR,
                "description": "Connect to server error."}}

        try:
            result.raise_for_status()
        except requests.HTTPError as exc:
            return {"error": {"code": exc.response.status_code,
                              "description": six.text_type(exc)}}

        if not filter_flag:
            LOG.info('''
            Request URL: %(url)s,
            Call Method: %(method)s,
            Request Data: %(data)s,
            Response Data: %(res)s,
            Result Data: %(res_json)s''', {'url': url, 'method': method,
                                           'data': data, 'res': result,
                                           'res_json': result.json()})

        if json_flag:
            return result
        else:
            return result.json()

    def _assert_rest_result(self, result, err_str):
        if result.get('result') != 0:
            msg = (_('%(err)s\nresult: %(res)s.') % {'err': err_str,
                                                     'res': result})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def get_version(self):
        url = 'rest/version'
        self.session.headers.update({
            "Referer": self.address + constants.BASIC_URI
        })
        result = self.call(url=url, method='GET', get_version=True)
        self._assert_rest_result(result, _('Get version session error.'))
        if result.get("currentVersion"):
            self.version = result["currentVersion"]

    def login(self):
        self.get_version()
        url = '/sec/login'
        data = {"userName": self.user, "password": self.password}
        result = self.call(url, 'POST', data=data,
                           call_timeout=constants.LOGIN_SOCKET_TIMEOUT,
                           filter_flag=True, json_flag=True)
        self._assert_rest_result(result.json(), _('Login session error.'))
        self.token = result.headers['X-Auth-Token']

        self.session.headers.update({
            "x-auth-token": self.token
        })

    def logout(self):
        url = '/sec/logout'
        if self.address:
            result = self.call(url, 'POST')
            self._assert_rest_result(result, _('Logout session error.'))

    def keep_alive(self):
        url = '/sec/keepAlive'
        result = self.call(url, 'POST', filter_flag=True)

        if result.get('result') == constants.ERROR_UNAUTHORIZED:
            try:
                self.login()
            except Exception:
                LOG.error('The FusionStorage may have been powered off. '
                          'Power on the FusionStorage and then log in.')
                raise
        else:
            self._assert_rest_result(result, _('Keep alive session error.'))

    def query_pool_info(self, pool_id=None):
        pool_id = str(pool_id)
        if pool_id != 'None':
            url = '/storagePool' + '?poolId=' + pool_id
        else:
            url = '/storagePool'
        result = self.call(url, 'GET', filter_flag=True)
        self._assert_rest_result(result, _("Query pool session error."))
        return result['storagePools']

    def query_volume_by_name(self, vol_name):
        url = '/volume/queryByName?volName=' + vol_name
        result = self.call(url, 'GET')
        if result.get('errorCode') == constants.VOLUME_NOT_EXIST:
            return None
        self._assert_rest_result(
            result, _("Query volume by name session error"))
        return result.get('lunDetailInfo')

    def query_volume_by_id(self, vol_id):
        url = '/volume/queryById?volId=' + vol_id
        result = self.call(url, 'GET')
        if result.get('errorCode') == constants.VOLUME_NOT_EXIST:
            return None
        self._assert_rest_result(
            result, _("Query volume by ID session error"))
        return result.get('lunDetailInfo')

    def create_volume(self, vol_name, vol_size, pool_id):
        url = '/volume/create'
        params = {"volName": vol_name, "volSize": vol_size, "poolId": pool_id}
        result = self.call(url, "POST", params)
        self._assert_rest_result(result, _('Create volume session error.'))

    def delete_volume(self, vol_name):
        url = '/volume/delete'
        params = {"volNames": [vol_name]}
        result = self.call(url, "POST", params)
        self._assert_rest_result(result, _('Delete volume session error.'))

    def attach_volume(self, vol_name, manage_ip):
        url = '/volume/attach'
        params = {"volName": [vol_name], "ipList": [manage_ip]}
        result = self.call(url, "POST", params)
        self._assert_rest_result(result, _('Attach volume session error.'))

        if int(result[vol_name][0]['errorCode']) != 0:
            msg = _("Host attach volume failed!")
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        return result

    def detach_volume(self, vol_name, manage_ip):
        url = '/volume/detach/'
        params = {"volName": [vol_name], "ipList": [manage_ip]}
        result = self.call(url, "POST", params)
        self._assert_rest_result(result, _('Detach volume session error.'))

    def expand_volume(self, vol_name, new_vol_size):
        url = '/volume/expand'
        params = {"volName": vol_name, "newVolSize": new_vol_size}
        result = self.call(url, "POST", params)
        self._assert_rest_result(result, _('Expand volume session error.'))

    def query_snapshot_by_name(self, pool_id, snapshot_name, page_num=1,
                               page_size=1000):
        # Filter the snapshot according to the name, while the "page_num" and
        # "page_size" must be set while using the interface.
        url = '/snapshot/list'
        params = {"poolId": pool_id, "pageNum": page_num,
                  "pageSize": page_size,
                  "filters": {"volumeName": snapshot_name}}

        result = self.call(url, "POST", params)
        self._assert_rest_result(
            result, _('query snapshot list session error.'))
        return result

    def create_snapshot(self, snapshot_name, vol_name):
        url = '/snapshot/create/'
        params = {"volName": vol_name, "snapshotName": snapshot_name}
        result = self.call(url, "POST", params)
        self._assert_rest_result(result, _('Create snapshot error.'))

    def delete_snapshot(self, snapshot_name):
        url = '/snapshot/delete/'
        params = {"snapshotName": snapshot_name}
        result = self.call(url, "POST", params)
        self._assert_rest_result(result, _('Delete snapshot session error.'))

    def create_volume_from_snapshot(self, snapshot_name, vol_name, vol_size):
        url = '/snapshot/volume/create/'
        params = {"src": snapshot_name, "volName": vol_name,
                  "volSize": vol_size}
        result = self.call(url, "POST", params)
        self._assert_rest_result(
            result, _('create volume from snapshot session error.'))

    def create_volume_from_volume(self, vol_name, vol_size, src_vol_name):
        temp_snapshot_name = "temp" + src_vol_name + "clone" + vol_name

        self.create_snapshot(vol_name=src_vol_name,
                             snapshot_name=temp_snapshot_name)

        self.create_volume_from_snapshot(snapshot_name=temp_snapshot_name,
                                         vol_name=vol_name, vol_size=vol_size)

        self.delete_snapshot(snapshot_name=temp_snapshot_name)



# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# Copyright (c) 2014 TrilioData, Inc
# Copyright (c) 2015 EMC Corporation
# Copyright (C) 2015 Kevin Fox <kevin@efox.cc>
# Copyright (C) 2015 Tom Barron <tpb@dyncloud.net>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Generic base class to implement metadata, compression and chunked data
   operations
"""

import abc
import hashlib
import json
import os
import sys

import eventlet
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import excutils
from oslo_utils import units
import six

from cinder.backup import driver
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import fields
from cinder.volume import volume_utils

if sys.platform == 'win32':
    from os_win import utilsfactory as os_win_utilsfactory

LOG = logging.getLogger(__name__)

backup_opts = [
    cfg.StrOpt('backup_compression_algorithm',
               default='zlib',
               ignore_case=True,
               choices=['none', 'off', 'no',
                        'zlib', 'gzip',
                        'bz2', 'bzip2'],
               help='Compression algorithm ("none" to disable)'),
]

CONF = cfg.CONF
CONF.register_opts(backup_opts)


# Object writer and reader returned by inheriting classes must not have any
# logging calls, as well as the compression libraries, as eventlet has a bug
# (https://github.com/eventlet/eventlet/issues/432) that would result in
# failures.

@six.add_metaclass(abc.ABCMeta)
class ChunkedBackupDriver(driver.BackupDriver):
    """Abstract chunked backup driver.

       Implements common functionality for backup drivers that store volume
       data in multiple "chunks" in a backup repository when the size of
       the backed up cinder volume exceeds the size of a backup repository
       "chunk."

       Provides abstract methods to be implemented in concrete chunking
       drivers.
    """

    DRIVER_VERSION = '1.0.0'
    DRIVER_VERSION_MAPPING = {'1.0.0': '_restore_v1'}

    def _get_compressor(self, algorithm):
        try:
            if algorithm.lower() in ('none', 'off', 'no'):
                return None
            if algorithm.lower() in ('zlib', 'gzip'):
                import zlib as compressor
                result = compressor
            elif algorithm.lower() in ('bz2', 'bzip2'):
                import bz2 as compressor
                result = compressor
            else:
                result = None
            if result:
                # NOTE(geguileo): Compression/Decompression starves
                # greenthreads so we use a native thread instead.
                return eventlet.tpool.Proxy(result)
        except ImportError:
            pass

        err = _('unsupported compression algorithm: %s') % algorithm
        raise ValueError(err)

    def __init__(self, context, chunk_size_bytes, sha_block_size_bytes,
                 backup_default_container, enable_progress_timer,
                 db=None):
        super(ChunkedBackupDriver, self).__init__(context, db)
        self.chunk_size_bytes = chunk_size_bytes
        self.sha_block_size_bytes = sha_block_size_bytes
        self.backup_default_container = backup_default_container
        self.enable_progress_timer = enable_progress_timer

        self.backup_timer_interval = CONF.backup_timer_interval
        self.data_block_num = CONF.backup_object_number_per_notification
        self.az = CONF.storage_availability_zone
        self.backup_compression_algorithm = CONF.backup_compression_algorithm
        self.compressor = \
            self._get_compressor(CONF.backup_compression_algorithm)
        self.support_force_delete = True

        if sys.platform == 'win32' and self.chunk_size_bytes % 4096:
            # The chunk size must be a multiple of the sector size. In order
            # to fail out early and avoid attaching the disks, we'll just
            # enforce the chunk size to be a multiple of 4096.
            err = _("Invalid chunk size. It must be a multiple of 4096.")
            raise exception.InvalidConfigurationValue(message=err)

    def _get_object_writer(self, container, object_name, extra_metadata=None):
        """Return writer proxy-wrapped to execute methods in native thread."""
        writer = self.get_object_writer(container, object_name, extra_metadata)
        return eventlet.tpool.Proxy(writer)

    def _get_object_reader(self, container, object_name, extra_metadata=None):
        """Return reader proxy-wrapped to execute methods in native thread."""
        reader = self.get_object_reader(container, object_name, extra_metadata)
        return eventlet.tpool.Proxy(reader)

    # To create your own "chunked" backup driver, implement the following
    # abstract methods.

    @abc.abstractmethod
    def put_container(self, container):
        """Create the container if needed. No failure if it pre-exists."""
        return

    @abc.abstractmethod
    def get_container_entries(self, container, prefix):
        """Get container entry names."""
        return

    @abc.abstractmethod
    def get_object_writer(self, container, object_name, extra_metadata=None):
        """Returns a writer object which stores the chunk data.

        The object returned should be a context handler that can be used in a
        "with" context.

        The object writer methods must not have any logging calls, as eventlet
        has a bug (https://github.com/eventlet/eventlet/issues/432) that would
        result in failures.
        """
        return

    @abc.abstractmethod
    def get_object_reader(self, container, object_name, extra_metadata=None):
        """Returns a reader object for the backed up chunk.

        The object reader methods must not have any logging calls, as eventlet
        has a bug (https://github.com/eventlet/eventlet/issues/432) that would
        result in failures.
        """
        return

    @abc.abstractmethod
    def delete_object(self, container, object_name):
        """Delete object from container."""
        return

    @abc.abstractmethod
    def _generate_object_name_prefix(self, backup):
        return

    @abc.abstractmethod
    def update_container_name(self, backup, container):
        """Allow sub-classes to override container name.

        This method exists so that sub-classes can override the container name
        as it comes in to the driver in the backup object. Implementations
        should return None if no change to the container name is desired.
        """
        return

    @abc.abstractmethod
    def get_extra_metadata(self, backup, volume):
        """Return extra metadata to use in prepare_backup.

        This method allows for collection of extra metadata in prepare_backup()
        which will be passed to get_object_reader() and get_object_writer().
        Subclass extensions can use this extra information to optimize
        data transfers. Return a json serializable object.
        """
        return

    def _create_container(self, backup):
        # Container's name will be decided by the driver (returned by method
        # update_container_name), if no change is required by the driver then
        # we'll use the one the backup object already has, but if it doesn't
        # have one backup_default_container will be used.
        new_container = self.update_container_name(backup, backup.container)
        if new_container:
            # If the driver is not really changing the name we don't want to
            # dirty the field in the object and save it to the DB with the same
            # value.
            if new_container != backup.container:
                backup.container = new_container
        elif backup.container is None:
            backup.container = self.backup_default_container

        LOG.debug('_create_container started, container: %(container)s,'
                  'backup: %(backup_id)s.',
                  {'container': backup.container, 'backup_id': backup.id})

        backup.save()
        self.put_container(backup.container)
        return backup.container

    def _generate_object_names(self, backup):
        prefix = backup['service_metadata']
        object_names = self.get_container_entries(backup['container'], prefix)
        LOG.debug('generated object list: %s.', object_names)
        return object_names

    def _metadata_filename(self, backup):
        object_name = backup['service_metadata']
        filename = '%s_metadata' % object_name
        return filename

    def _sha256_filename(self, backup):
        object_name = backup['service_metadata']
        filename = '%s_sha256file' % object_name
        return filename

    def _write_metadata(self, backup, volume_id, container, object_list,
                        volume_meta, extra_metadata=None):
        filename = self._metadata_filename(backup)
        LOG.debug('_write_metadata started, container name: %(container)s,'
                  ' metadata filename: %(filename)s.',
                  {'container': container, 'filename': filename})
        metadata = {}
        metadata['version'] = self.DRIVER_VERSION
        metadata['backup_id'] = backup['id']
        metadata['volume_id'] = volume_id
        metadata['backup_name'] = backup['display_name']
        metadata['backup_description'] = backup['display_description']
        metadata['created_at'] = str(backup['created_at'])
        metadata['objects'] = object_list
        metadata['parent_id'] = backup['parent_id']
        metadata['volume_meta'] = volume_meta
        if extra_metadata:
            metadata['extra_metadata'] = extra_metadata
        metadata_json = json.dumps(metadata, sort_keys=True, indent=2)
        if six.PY3:
            metadata_json = metadata_json.encode('utf-8')
        with self._get_object_writer(container, filename) as writer:
            writer.write(metadata_json)
        LOG.debug('_write_metadata finished. Metadata: %s.', metadata_json)

    def _write_sha256file(self, backup, volume_id, container, sha256_list):
        filename = self._sha256_filename(backup)
        LOG.debug('_write_sha256file started, container name: %(container)s,'
                  ' sha256file filename: %(filename)s.',
                  {'container': container, 'filename': filename})
        sha256file = {}
        sha256file['version'] = self.DRIVER_VERSION
        sha256file['backup_id'] = backup['id']
        sha256file['volume_id'] = volume_id
        sha256file['backup_name'] = backup['display_name']
        sha256file['backup_description'] = backup['display_description']
        sha256file['created_at'] = six.text_type(backup['created_at'])
        sha256file['chunk_size'] = self.sha_block_size_bytes
        sha256file['sha256s'] = sha256_list
        sha256file_json = json.dumps(sha256file, sort_keys=True, indent=2)
        if six.PY3:
            sha256file_json = sha256file_json.encode('utf-8')
        with self._get_object_writer(container, filename) as writer:
            writer.write(sha256file_json)
        LOG.debug('_write_sha256file finished.')

    def _read_metadata(self, backup):
        container = backup['container']
        filename = self._metadata_filename(backup)
        LOG.debug('_read_metadata started, container name: %(container)s, '
                  'metadata filename: %(filename)s.',
                  {'container': container, 'filename': filename})
        with self._get_object_reader(container, filename) as reader:
            metadata_json = reader.read()
        if six.PY3:
            metadata_json = metadata_json.decode('utf-8')
        metadata = json.loads(metadata_json)
        LOG.debug('_read_metadata finished. Metadata: %s.', metadata_json)
        return metadata

    def _read_sha256file(self, backup):
        container = backup['container']
        filename = self._sha256_filename(backup)
        LOG.debug('_read_sha256file started, container name: %(container)s, '
                  'sha256 filename: %(filename)s.',
                  {'container': container, 'filename': filename})
        with self._get_object_reader(container, filename) as reader:
            sha256file_json = reader.read()
        if six.PY3:
            sha256file_json = sha256file_json.decode('utf-8')
        sha256file = json.loads(sha256file_json)
        LOG.debug('_read_sha256file finished.')
        return sha256file

    def _prepare_backup(self, backup):
        """Prepare the backup process and return the backup metadata."""
        volume = self.db.volume_get(self.context, backup.volume_id)

        if volume['size'] <= 0:
            err = _('volume size %d is invalid.') % volume['size']
            raise exception.InvalidVolume(reason=err)

        container = self._create_container(backup)

        object_prefix = self._generate_object_name_prefix(backup)
        backup.service_metadata = object_prefix
        backup.save()

        volume_size_bytes = volume['size'] * units.Gi
        availability_zone = self.az
        LOG.debug('starting backup of volume: %(volume_id)s,'
                  ' volume size: %(volume_size_bytes)d, object names'
                  ' prefix %(object_prefix)s, availability zone:'
                  ' %(availability_zone)s',
                  {
                      'volume_id': backup.volume_id,
                      'volume_size_bytes': volume_size_bytes,
                      'object_prefix': object_prefix,
                      'availability_zone': availability_zone,
                  })
        object_meta = {'id': 1, 'list': [], 'prefix': object_prefix,
                       'volume_meta': None}
        object_sha256 = {'id': 1, 'sha256s': [], 'prefix': object_prefix}
        extra_metadata = self.get_extra_metadata(backup, volume)
        if extra_metadata is not None:
            object_meta['extra_metadata'] = extra_metadata

        return (object_meta, object_sha256, extra_metadata, container,
                volume_size_bytes)

    def _backup_chunk(self, backup, container, data, data_offset,
                      object_meta, extra_metadata):
        """Backup data chunk based on the object metadata and offset."""
        object_prefix = object_meta['prefix']
        object_list = object_meta['list']

        object_id = object_meta['id']
        object_name = '%s-%05d' % (object_prefix, object_id)
        obj = {}
        obj[object_name] = {}
        obj[object_name]['offset'] = data_offset
        obj[object_name]['length'] = len(data)
        LOG.debug('Backing up chunk of data from volume.')
        algorithm, output_data = self._prepare_output_data(data)
        obj[object_name]['compression'] = algorithm
        LOG.debug('About to put_object')
        with self._get_object_writer(
                container, object_name, extra_metadata=extra_metadata
        ) as writer:
            writer.write(output_data)
        md5 = eventlet.tpool.execute(hashlib.md5, data).hexdigest()
        obj[object_name]['md5'] = md5
        LOG.debug('backup MD5 for %(object_name)s: %(md5)s',
                  {'object_name': object_name, 'md5': md5})
        object_list.append(obj)
        object_id += 1
        object_meta['list'] = object_list
        object_meta['id'] = object_id

        LOG.debug('Calling eventlet.sleep(0)')
        eventlet.sleep(0)

    def _prepare_output_data(self, data):
        if self.compressor is None:
            return 'none', data
        data_size_bytes = len(data)
        # Execute compression in native thread so it doesn't prevent
        # cooperative greenthread switching.
        compressed_data = self.compressor.compress(data)
        comp_size_bytes = len(compressed_data)
        algorithm = CONF.backup_compression_algorithm.lower()
        if comp_size_bytes >= data_size_bytes:
            LOG.debug('Compression of this chunk was ineffective: '
                      'original length: %(data_size_bytes)d, '
                      'compressed length: %(compressed_size_bytes)d. '
                      'Using original data for this chunk.',
                      {'data_size_bytes': data_size_bytes,
                       'compressed_size_bytes': comp_size_bytes,
                       })
            return 'none', data
        LOG.debug('Compressed %(data_size_bytes)d bytes of data '
                  'to %(comp_size_bytes)d bytes using %(algorithm)s.',
                  {'data_size_bytes': data_size_bytes,
                   'comp_size_bytes': comp_size_bytes,
                   'algorithm': algorithm,
                   })
        return algorithm, compressed_data

    def _finalize_backup(self, backup, container, object_meta, object_sha256):
        """Write the backup's metadata to the backup repository."""
        object_list = object_meta['list']
        object_id = object_meta['id']
        volume_meta = object_meta['volume_meta']
        sha256_list = object_sha256['sha256s']
        extra_metadata = object_meta.get('extra_metadata')
        self._write_sha256file(backup,
                               backup.volume_id,
                               container,
                               sha256_list)
        self._write_metadata(backup,
                             backup.volume_id,
                             container,
                             object_list,
                             volume_meta,
                             extra_metadata)
        # NOTE(whoami-rajat) : The object_id variable is used to name
        # the backup objects and hence differs from the object_count
        # variable, therefore the increment of object_id value in the last
        # iteration of _backup_chunk() method shouldn't be reflected in the
        # object_count variable.
        backup.object_count = object_id - 1
        backup.save()
        LOG.debug('backup %s finished.', backup['id'])

    def _backup_metadata(self, backup, object_meta):
        """Backup volume metadata.

        NOTE(dosaboy): the metadata we are backing up is obtained from a
                       versioned api so we should not alter it in any way here.
                       We must also be sure that the service that will perform
                       the restore is compatible with version used.
        """
        json_meta = self.get_metadata(backup['volume_id'])
        if not json_meta:
            LOG.debug("No volume metadata to backup.")
            return

        object_meta["volume_meta"] = json_meta

    def _send_progress_end(self, context, backup, object_meta):
        object_meta['backup_percent'] = 100
        volume_utils.notify_about_backup_usage(context,
                                               backup,
                                               "createprogress",
                                               extra_usage_info=
                                               object_meta)

    def _send_progress_notification(self, context, backup, object_meta,
                                    total_block_sent_num, total_volume_size):
        backup_percent = total_block_sent_num * 100 / total_volume_size
        object_meta['backup_percent'] = backup_percent
        volume_utils.notify_about_backup_usage(context,
                                               backup,
                                               "createprogress",
                                               extra_usage_info=
                                               object_meta)

    def _get_win32_phys_disk_size(self, disk_path):
        win32_diskutils = os_win_utilsfactory.get_diskutils()
        disk_number = win32_diskutils.get_device_number_from_device_name(
            disk_path)
        return win32_diskutils.get_disk_size(disk_number)

    def _calculate_sha(self, data):
        """Calculate SHA256 of a data chunk.

        This method cannot log anything as it is called on a native thread.
        """
        # NOTE(geguileo): Using memoryview to avoid data copying when slicing
        # for the sha256 call.
        chunk = memoryview(data)
        shalist = []
        off = 0
        datalen = len(chunk)
        while off < datalen:
            chunk_end = min(datalen, off + self.sha_block_size_bytes)
            block = chunk[off:chunk_end]
            sha = hashlib.sha256(block).hexdigest()
            shalist.append(sha)
            off += self.sha_block_size_bytes
        return shalist

    def backup(self, backup, volume_file, backup_metadata=True):
        """Backup the given volume.

           If backup['parent_id'] is given, then an incremental backup
           is performed.
        """
        if self.chunk_size_bytes % self.sha_block_size_bytes:
            err = _('Chunk size is not multiple of '
                    'block size for creating hash.')
            raise exception.InvalidBackup(reason=err)

        # Read the shafile of the parent backup if backup['parent_id']
        # is given.
        parent_backup_shafile = None
        parent_backup = None
        if backup.parent_id:
            parent_backup = objects.Backup.get_by_id(self.context,
                                                     backup.parent_id)
            parent_backup_shafile = self._read_sha256file(parent_backup)
            parent_backup_shalist = parent_backup_shafile['sha256s']
            if (parent_backup_shafile['chunk_size'] !=
                    self.sha_block_size_bytes):
                err = (_('Hash block size has changed since the last '
                         'backup. New hash block size: %(new)s. Old hash '
                         'block size: %(old)s. Do a full backup.')
                       % {'old': parent_backup_shafile['chunk_size'],
                          'new': self.sha_block_size_bytes})
                raise exception.InvalidBackup(reason=err)
            # If the volume size increased since the last backup, fail
            # the incremental backup and ask user to do a full backup.
            if backup.size > parent_backup.size:
                err = _('Volume size increased since the last '
                        'backup. Do a full backup.')
                raise exception.InvalidBackup(reason=err)

        if sys.platform == 'win32':
            # When dealing with Windows physical disks, we need the exact
            # size of the disk. Attempting to read passed this boundary will
            # lead to an IOError exception. At the same time, we cannot
            # seek to the end of file.
            win32_disk_size = self._get_win32_phys_disk_size(volume_file.name)

        (object_meta, object_sha256, extra_metadata, container,
         volume_size_bytes) = self._prepare_backup(backup)

        counter = 0
        total_block_sent_num = 0

        # There are two mechanisms to send the progress notification.
        # 1. The notifications are periodically sent in a certain interval.
        # 2. The notifications are sent after a certain number of chunks.
        # Both of them are working simultaneously during the volume backup,
        # when "chunked" backup drivers are deployed.
        def _notify_progress():
            self._send_progress_notification(self.context, backup,
                                             object_meta,
                                             total_block_sent_num,
                                             volume_size_bytes)
        timer = loopingcall.FixedIntervalLoopingCall(
            _notify_progress)
        if self.enable_progress_timer:
            timer.start(interval=self.backup_timer_interval)

        sha256_list = object_sha256['sha256s']
        shaindex = 0
        is_backup_canceled = False
        while True:
            # First of all, we check the status of this backup. If it
            # has been changed to delete or has been deleted, we cancel the
            # backup process to do forcing delete.
            with backup.as_read_deleted():
                backup.refresh()
            if backup.status in (fields.BackupStatus.DELETING,
                                 fields.BackupStatus.DELETED):
                is_backup_canceled = True
                # To avoid the chunk left when deletion complete, need to
                # clean up the object of chunk again.
                self.delete_backup(backup)
                LOG.debug('Cancel the backup process of %s.', backup.id)
                break
            data_offset = volume_file.tell()

            if sys.platform == 'win32':
                read_bytes = min(self.chunk_size_bytes,
                                 win32_disk_size - data_offset)
            else:
                read_bytes = self.chunk_size_bytes
            data = volume_file.read(read_bytes)

            if data == b'':
                break

            # Calculate new shas with the datablock.
            shalist = eventlet.tpool.execute(self._calculate_sha, data)
            sha256_list.extend(shalist)

            # If parent_backup is not None, that means an incremental
            # backup will be performed.
            if parent_backup:
                # Find the extent that needs to be backed up.
                extent_off = -1
                for idx, sha in enumerate(shalist):
                    if sha != parent_backup_shalist[shaindex]:
                        if extent_off == -1:
                            # Start of new extent.
                            extent_off = idx * self.sha_block_size_bytes
                    else:
                        if extent_off != -1:
                            # We've reached the end of extent.
                            extent_end = idx * self.sha_block_size_bytes
                            segment = data[extent_off:extent_end]
                            self._backup_chunk(backup, container, segment,
                                               data_offset + extent_off,
                                               object_meta,
                                               extra_metadata)
                            extent_off = -1
                    shaindex += 1

                # The last extent extends to the end of data buffer.
                if extent_off != -1:
                    extent_end = len(data)
                    segment = data[extent_off:extent_end]
                    self._backup_chunk(backup, container, segment,
                                       data_offset + extent_off,
                                       object_meta, extra_metadata)
                    extent_off = -1
            else:  # Do a full backup.
                self._backup_chunk(backup, container, data, data_offset,
                                   object_meta, extra_metadata)

            # Notifications
            total_block_sent_num += self.data_block_num
            counter += 1
            if counter == self.data_block_num:
                # Send the notification to Ceilometer when the chunk
                # number reaches the data_block_num.  The backup percentage
                # is put in the metadata as the extra information.
                self._send_progress_notification(self.context, backup,
                                                 object_meta,
                                                 total_block_sent_num,
                                                 volume_size_bytes)
                # Reset the counter
                counter = 0

        # Stop the timer.
        timer.stop()
        # If backup has been cancelled we have nothing more to do
        # but timer.stop().
        if is_backup_canceled:
            return
        # All the data have been sent, the backup_percent reaches 100.
        self._send_progress_end(self.context, backup, object_meta)

        object_sha256['sha256s'] = sha256_list
        if backup_metadata:
            try:
                self._backup_metadata(backup, object_meta)
            # Whatever goes wrong, we want to log, cleanup, and re-raise.
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.exception("Backup volume metadata failed.")
                    self.delete_backup(backup)

        self._finalize_backup(backup, container, object_meta, object_sha256)

    def _restore_v1(self, backup, volume_id, metadata, volume_file,
                    requested_backup):
        """Restore a v1 volume backup.

        Raises BackupRestoreCancel on any requested_backup status change, we
        ignore the backup parameter for this check since that's only the
        current data source from the list of backup sources.
        """
        backup_id = backup['id']
        LOG.debug('v1 volume backup restore of %s started.', backup_id)
        extra_metadata = metadata.get('extra_metadata')
        container = backup['container']
        metadata_objects = metadata['objects']
        metadata_object_names = []
        for obj in metadata_objects:
            metadata_object_names.extend(obj.keys())
        LOG.debug('metadata_object_names = %s.', metadata_object_names)
        prune_list = [self._metadata_filename(backup),
                      self._sha256_filename(backup)]
        object_names = [object_name for object_name in
                        self._generate_object_names(backup)
                        if object_name not in prune_list]
        if sorted(object_names) != sorted(metadata_object_names):
            err = _('restore_backup aborted, actual object list '
                    'does not match object list stored in metadata.')
            raise exception.InvalidBackup(reason=err)

        for metadata_object in metadata_objects:
            # Abort when status changes to error, available, or anything else
            with requested_backup.as_read_deleted():
                requested_backup.refresh()
            if requested_backup.status != fields.BackupStatus.RESTORING:
                raise exception.BackupRestoreCancel(back_id=backup.id,
                                                    vol_id=volume_id)

            object_name, obj = list(metadata_object.items())[0]
            LOG.debug('restoring object. backup: %(backup_id)s, '
                      'container: %(container)s, object name: '
                      '%(object_name)s, volume: %(volume_id)s.',
                      {
                          'backup_id': backup_id,
                          'container': container,
                          'object_name': object_name,
                          'volume_id': volume_id,
                      })

            with self._get_object_reader(
                    container, object_name,
                    extra_metadata=extra_metadata) as reader:
                body = reader.read()
            compression_algorithm = metadata_object[object_name]['compression']
            decompressor = self._get_compressor(compression_algorithm)
            volume_file.seek(obj['offset'])
            if decompressor is not None:
                LOG.debug('decompressing data using %s algorithm',
                          compression_algorithm)
                decompressed = decompressor.decompress(body)
                body = None  # Allow Python to free it
                volume_file.write(decompressed)
                decompressed = None  # Allow Python to free it
            else:
                volume_file.write(body)
                body = None  # Allow Python to free it

            # force flush every write to avoid long blocking write on close
            volume_file.flush()

            # Be tolerant to IO implementations that do not support fileno()
            try:
                fileno = volume_file.fileno()
            except IOError:
                LOG.info("volume_file does not support fileno() so skipping "
                         "fsync()")
            else:
                os.fsync(fileno)

            # Restoring a backup to a volume can take some time. Yield so other
            # threads can run, allowing for among other things the service
            # status to be updated
            eventlet.sleep(0)
        LOG.debug('v1 volume backup restore of %s finished.',
                  backup_id)

    def restore(self, backup, volume_id, volume_file):
        """Restore the given volume backup from backup repository.

        Raises BackupRestoreCancel on any backup status change.
        """
        backup_id = backup['id']
        container = backup['container']
        object_prefix = backup['service_metadata']
        LOG.debug('starting restore of backup %(object_prefix)s '
                  'container: %(container)s, to volume %(volume_id)s, '
                  'backup: %(backup_id)s.',
                  {
                      'object_prefix': object_prefix,
                      'container': container,
                      'volume_id': volume_id,
                      'backup_id': backup_id,
                  })
        metadata = self._read_metadata(backup)
        metadata_version = metadata['version']
        LOG.debug('Restoring backup version %s', metadata_version)
        try:
            restore_func = getattr(self, self.DRIVER_VERSION_MAPPING.get(
                metadata_version))
        except TypeError:
            err = (_('No support to restore backup version %s')
                   % metadata_version)
            raise exception.InvalidBackup(reason=err)

        # Build a list of backups based on parent_id. A full backup
        # will be the last one in the list.
        backup_list = []
        backup_list.append(backup)
        current_backup = backup
        while current_backup.parent_id:
            prev_backup = objects.Backup.get_by_id(self.context,
                                                   current_backup.parent_id)
            backup_list.append(prev_backup)
            current_backup = prev_backup

        # Do a full restore first, then layer the incremental backups
        # on top of it in order.
        index = len(backup_list) - 1
        while index >= 0:
            backup1 = backup_list[index]
            index = index - 1
            metadata = self._read_metadata(backup1)
            restore_func(backup1, volume_id, metadata, volume_file, backup)

            volume_meta = metadata.get('volume_meta', None)
            try:
                if volume_meta:
                    self.put_metadata(volume_id, volume_meta)
                else:
                    LOG.debug("No volume metadata in this backup.")
            except exception.BackupMetadataUnsupportedVersion:
                msg = _("Metadata restore failed due to incompatible version.")
                LOG.error(msg)
                raise exception.BackupOperationError(msg)

        LOG.debug('restore %(backup_id)s to %(volume_id)s finished.',
                  {'backup_id': backup_id, 'volume_id': volume_id})

    def delete_backup(self, backup):
        """Delete the given backup."""
        container = backup['container']
        object_prefix = backup['service_metadata']
        LOG.debug('delete started, backup: %(id)s, container: %(cont)s, '
                  'prefix: %(pre)s.',
                  {'id': backup['id'],
                   'cont': container,
                   'pre': object_prefix})

        if container is not None and object_prefix is not None:
            object_names = []
            try:
                object_names = self._generate_object_names(backup)
            except Exception:
                LOG.warning('Error while listing objects, continuing'
                            ' with delete.')

            for object_name in object_names:
                self.delete_object(container, object_name)
                LOG.debug('deleted object: %(object_name)s'
                          ' in container: %(container)s.',
                          {
                              'object_name': object_name,
                              'container': container
                          })
                # Deleting a backup's objects can take some time.
                # Yield so other threads can run
                eventlet.sleep(0)

        LOG.debug('delete %s finished.', backup['id'])



# Copyright (c) 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for GlusterFS backup driver.

"""
import os

import mock
from os_brick.remotefs import remotefs as remotefs_brick

from cinder.backup.drivers import glusterfs
from cinder import context
from cinder import exception
from cinder import test
from cinder import utils

FAKE_BACKUP_MOUNT_POINT_BASE = '/fake/mount-point-base'
FAKE_HOST = 'fake_host'
FAKE_VOL_NAME = 'backup_vol'
FAKE_BACKUP_SHARE = '%s:%s' % (FAKE_HOST, FAKE_VOL_NAME)
FAKE_BACKUP_PATH = os.path.join(FAKE_BACKUP_MOUNT_POINT_BASE,
                                'e51e43e3c63fd5770e90e58e2eafc709')


class BackupGlusterfsShareTestCase(test.TestCase):

    def setUp(self):
        super(BackupGlusterfsShareTestCase, self).setUp()
        self.ctxt = context.get_admin_context()

    def test_check_configuration(self):
        self.override_config('glusterfs_backup_share', FAKE_BACKUP_SHARE)
        self.mock_object(glusterfs.GlusterfsBackupDriver,
                         '_init_backup_repo_path',
                         return_value=FAKE_BACKUP_PATH)

        driver = glusterfs.GlusterfsBackupDriver(self.ctxt)
        driver.check_for_setup_error()

    def test_check_configuration_no_backup_share(self):
        self.override_config('glusterfs_backup_share', None)
        self.mock_object(glusterfs.GlusterfsBackupDriver,
                         '_init_backup_repo_path',
                         return_value=FAKE_BACKUP_PATH)

        driver = glusterfs.GlusterfsBackupDriver(self.ctxt)
        self.assertRaises(exception.InvalidConfigurationValue,
                          driver.check_for_setup_error)

    def test_init_backup_repo_path(self):
        self.override_config('glusterfs_backup_share', FAKE_BACKUP_SHARE)
        self.override_config('glusterfs_backup_mount_point',
                             FAKE_BACKUP_MOUNT_POINT_BASE)
        mock_remotefsclient = mock.Mock()
        mock_remotefsclient.get_mount_point = mock.Mock(
            return_value=FAKE_BACKUP_PATH)
        self.mock_object(glusterfs.GlusterfsBackupDriver,
                         'check_for_setup_error')
        self.mock_object(remotefs_brick, 'RemoteFsClient',
                         return_value=mock_remotefsclient)
        self.mock_object(os, 'getegid',
                         return_value=333333)
        self.mock_object(utils, 'get_file_gid',
                         return_value=333333)
        self.mock_object(utils, 'get_file_mode',
                         return_value=00000)
        self.mock_object(utils, 'get_root_helper')

        with mock.patch.object(glusterfs.GlusterfsBackupDriver,
                               '_init_backup_repo_path'):
            driver = glusterfs.GlusterfsBackupDriver(self.ctxt)
        self.mock_object(driver, '_execute')
        path = driver._init_backup_repo_path()

        self.assertEqual(FAKE_BACKUP_PATH, path)
        utils.get_root_helper.called_once()
        mock_remotefsclient.mount.assert_called_once_with(FAKE_BACKUP_SHARE)
        mock_remotefsclient.get_mount_point.assert_called_once_with(
            FAKE_BACKUP_SHARE)



# Copyright (c) 2019 MacroSAN Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Volume Drivers Config Registration documents for MacroSAN SAN."""

from oslo_config import cfg


macrosan_opts = [
    # sdas login_info
    cfg.ListOpt('macrosan_sdas_ipaddrs',
                help="MacroSAN sdas devices' ip addresses"),
    cfg.StrOpt('macrosan_sdas_username',
               help="MacroSAN sdas devices' username"),
    cfg.StrOpt('macrosan_sdas_password',
               secret=True,
               help="MacroSAN sdas devices' password"),
    # replication login_info
    cfg.ListOpt('macrosan_replication_ipaddrs',
                help="MacroSAN replication devices' ip addresses"),
    cfg.StrOpt('macrosan_replication_username',
               help="MacroSAN replication devices' username"),
    cfg.StrOpt('macrosan_replication_password',
               secret=True,
               help="MacroSAN replication devices' password"),
    cfg.ListOpt('macrosan_replication_destination_ports',
                sample_default="eth-1:0/eth-1:1, eth-2:0/eth-2:1",
                help="Slave device"),
    # device_features
    cfg.StrOpt('macrosan_pool', quotes=True,
               help='Pool to use for volume creation'),
    cfg.IntOpt('macrosan_thin_lun_extent_size',
               default=8,
               help="Set the thin lun's extent size"),
    cfg.IntOpt('macrosan_thin_lun_low_watermark',
               default=5,
               help="Set the thin lun's low watermark"),
    cfg.IntOpt('macrosan_thin_lun_high_watermark',
               default=20,
               help="Set the thin lun's high watermark"),
    cfg.BoolOpt('macrosan_force_unmap_itl',
                default=True,
                help="Force disconnect while deleting volume"),
    cfg.FloatOpt('macrosan_snapshot_resource_ratio',
                 default=1.0,
                 help="Set snapshot's resource ratio"),
    cfg.BoolOpt('macrosan_log_timing',
                default=True,
                help="Whether enable log timing"),
    # fc connection
    cfg.IntOpt('macrosan_fc_use_sp_port_nr',
               default=1,
               max=4,
               help="The use_sp_port_nr parameter is the number of "
                    "online FC ports used by the single-ended memory "
                    "when the FC connection is established in the switch "
                    "non-all-pass mode. The maximum is 4"),
    cfg.BoolOpt('macrosan_fc_keep_mapped_ports',
                default=True,
                help="In the case of an FC connection, the configuration "
                     "item associated with the port is maintained."),
    # iscsi connection
    cfg.ListOpt('macrosan_client',
                help="""Macrosan iscsi_clients list.
                You can configure multiple clients.
                You can configure it in this format:
                (host; client_name; sp1_iscsi_port; sp2_iscsi_port),
                (host; client_name; sp1_iscsi_port; sp2_iscsi_port)
                Important warning, Client_name has the following requirements:
                    [a-zA-Z0-9.-_:], the maximum number of characters is 31
                E.g:
                (controller1; device1; eth-1:0; eth-2:0),
                (controller2; device2; eth-1:0/eth-1:1; eth-2:0/eth-2:1),
                """),
    cfg.StrOpt('macrosan_client_default',
               help="This is the default connection ports' name for iscsi. "
                    "This default configuration is used "
                    "when no host related information is obtained."
                    "E.g: eth-1:0/eth-1:1; eth-2:0/eth-2:1")
]



# Copyright 2017 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import webob

from cinder import context
from cinder import db
from cinder import exception
from cinder import rpc
from cinder import test

from cinder.api import microversions as mv
from cinder.api.v3 import group_specs as v3_group_specs
from cinder.tests.unit.api import fakes
from cinder.tests.unit import fake_constants as fake

fake_group_specs = {
    'key1': 'value1',
    'key2': 'value2'
}

create_fake_group_specs = {
    'group_specs': {
        'key1': 'value1',
        'key2': 'value2'
    }
}

update_fake_group_specs = {
    'id': 'any_string'
}

incorrect_fake_group_specs = {
    'group_specs': {
        'key#': 'value1',
        'key2': 'value2'
    }
}


class GroupSpecsTestCase(test.TestCase):
    """test cases for the group specs API"""

    def setUp(self):
        super(GroupSpecsTestCase, self).setUp()
        self.controller = v3_group_specs.GroupTypeSpecsController()
        self.ctxt = context.RequestContext(
            user_id=fake.USER_ID,
            project_id=fake.PROJECT_ID,
            is_admin=True)

    @mock.patch.object(db, 'group_type_get', return_value={})
    @mock.patch.object(db, 'group_type_specs_get',
                       return_value=fake_group_specs)
    def test_group_types_index(self,
                               mock_group_type_specs_get,
                               mock_group_type_get):
        req = fakes.HTTPRequest.blank('v3/%s/group_specs' %
                                      fake.PROJECT_ID,
                                      use_admin_context=True,
                                      version=mv.GROUP_TYPE)
        req.environ['cinder.context'] = self.ctxt
        res_dict = self.controller.index(req, fake.GROUP_ID)
        group_specs_dict = res_dict['group_specs']
        mock_group_type_specs_get.assert_called()
        self.assertEqual('value1', group_specs_dict['key1'])
        self.assertEqual('value2', group_specs_dict['key2'])

    @mock.patch.object(rpc, 'get_notifier')
    @mock.patch.object(db, 'group_type_get', return_value={})
    @mock.patch.object(db, 'group_type_specs_update_or_create',
                       return_value={})
    def test_group_types_create(self,
                                mock_update_or_create,
                                mock_group_type_get,
                                mock_rpc_notifier):
        req = fakes.HTTPRequest.blank('v3/%s/group_specs' %
                                      fake.PROJECT_ID,
                                      use_admin_context=True,
                                      version=mv.GROUP_TYPE)
        self.controller.create(req, fake.GROUP_ID,
                               body=create_fake_group_specs)
        self.assertTrue(mock_rpc_notifier.called)

    @mock.patch.object(rpc, 'get_notifier')
    @mock.patch.object(db, 'group_type_get', return_value={})
    @mock.patch.object(db, 'group_type_specs_get',
                       return_value=fake_group_specs)
    @mock.patch.object(db, 'group_type_specs_update_or_create',
                       return_value={})
    def test_group_types_update(self,
                                mock_update_or_create,
                                mock_typ_specs_get,
                                mock_group_type_get,
                                mock_rpc_notifier):
        req = fakes.HTTPRequest.blank('v3/%s/group_specs' %
                                      fake.PROJECT_ID,
                                      use_admin_context=True,
                                      version=mv.GROUP_TYPE)
        self.controller.update(req,
                               fake.GROUP_TYPE_ID,
                               'id',
                               body=update_fake_group_specs)
        self.assertTrue(mock_rpc_notifier.called)

    @mock.patch.object(db, 'group_type_specs_get',
                       return_value=fake_group_specs)
    @mock.patch.object(db, 'group_type_get', return_value={})
    def test_group_types_show(self,
                              mock_group_type_get,
                              mock_fake_group_specs):
        req = fakes.HTTPRequest.blank('v3/%s/group_specs' %
                                      fake.PROJECT_ID,
                                      use_admin_context=True,
                                      version=mv.GROUP_TYPE)
        res_dict = self.controller.show(req, fake.GROUP_TYPE_ID, 'key1')
        self.assertEqual('value1', res_dict['key1'])

    @mock.patch.object(rpc, 'get_notifier')
    @mock.patch.object(db, 'group_type_specs_delete', return_value={})
    @mock.patch.object(db, 'group_type_get', return_value={})
    def test_group_types_delete(self,
                                mock_group_type_get,
                                mock_group_spec_delete,
                                rpc_notifier_mock):
        req = fakes.HTTPRequest.blank('v3/%s/group_specs' %
                                      fake.PROJECT_ID,
                                      use_admin_context=True,
                                      version=mv.GROUP_TYPE)
        self.controller.delete(req, fake.GROUP_TYPE_ID, 'key1')
        self.assertTrue(rpc_notifier_mock.called)

    @mock.patch.object(rpc, 'get_notifier')
    @mock.patch.object(db, 'group_type_specs_update_or_create',
                       return_value={})
    def test_check_type_should_raise_exception(self,
                                               mock_db_update_or_create,
                                               mock_rpc_notifier):
        req = fakes.HTTPRequest.blank('v3/%s/group_specs' %
                                      fake.PROJECT_ID,
                                      use_admin_context=True,
                                      version=mv.GROUP_TYPE)
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.create,
                          req,
                          fake.GROUP_ID,
                          body=create_fake_group_specs)

    @mock.patch.object(rpc, 'get_notifier')
    @mock.patch.object(db, 'group_type_get', return_value={})
    def test_delete_should_raise_exception(self,
                                           mock_group_type_get,
                                           mock_get_notifier):
        req = fakes.HTTPRequest.blank('v3/%s/group_specs' %
                                      fake.PROJECT_ID,
                                      use_admin_context=True,
                                      version=mv.GROUP_TYPE)
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.delete,
                          req,
                          fake.GROUP_TYPE_ID,
                          'key1')

    @mock.patch.object(db, 'group_type_get', return_value={})
    def test_update_should_raise_exceptions(self, mock_group_type_get):
        req = fakes.HTTPRequest.blank('v3/%s/group_specs' %
                                      fake.PROJECT_ID,
                                      use_admin_context=True,
                                      version=mv.GROUP_TYPE)
        self.assertRaises(exception.ValidationError,
                          self.controller.update,
                          req,
                          fake.GROUP_TYPE_ID,
                          'id', body=None)
        self.assertRaises(exception.ValidationError, self.controller.update,
                          req, fake.GROUP_TYPE_ID, 'id',
                          body=fake_group_specs)
        self.assertRaises(exception.ValidationError, self.controller.update,
                          req, fake.GROUP_TYPE_ID, 'key1',
                          body=fake_group_specs)

    @mock.patch.object(db, 'group_type_specs_get',
                       return_value=fake_group_specs)
    @mock.patch.object(db, 'group_type_get', return_value={})
    def test_show_should_raise_exception(self,
                                         mock_group_type_get,
                                         mock_group_type_specs_get):
        req = fakes.HTTPRequest.blank('v3/%s/group_specs' %
                                      fake.PROJECT_ID,
                                      use_admin_context=True,
                                      version=mv.GROUP_TYPE)
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.show,
                          req,
                          fake.GROUP_TYPE_ID,
                          'key')

    @mock.patch.object(rpc, 'get_notifier')
    @mock.patch.object(db, 'group_type_get', return_value={})
    @mock.patch.object(db, 'group_type_specs_update_or_create',
                       return_value={})
    def test_check_key_name_should_raise_exception(self,
                                                   mock_update_or_create,
                                                   mock_group_type_get,
                                                   mock_rpc_notifier):
        req = fakes.HTTPRequest.blank('v3/%s/group_specs' %
                                      fake.PROJECT_ID,
                                      use_admin_context=True,
                                      version=mv.GROUP_TYPE)
        self.assertRaises(exception.ValidationError, self.controller.create,
                          req, fake.GROUP_ID, body=incorrect_fake_group_specs)



# Copyright (c) 2017-2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import datetime
import platform
import prettytable
import six
import time
import traceback
import types

from oslo_log import log as logging

from cinder.objects import volume
from cinder import version
from cinder.volume.drivers.dell_emc.powermax import utils
LOG = logging.getLogger(__name__)
CLEANUP_LIST = ['masking_view', 'child_storage_group', 'parent_storage_group',
                'initiator_group', 'port_group', 'storage_group']


def debug_required(func):
    """Only execute the function if debug is enabled."""

    def func_wrapper(*args, **kwargs):
        try:
            if args[0].is_debug:
                return func(*args, **kwargs)
            else:
                pass
        except Exception:
            LOG.warning("Volume metadata logging failure. "
                        "Exception is %s.", traceback.format_exc())

    return func_wrapper


class PowerMaxVolumeMetadata(object):
    """Gathers PowerMax/VMAX specific volume information.

    Also gathers Unisphere, Microcode OS/distribution and python versions.
    """

    def __init__(self, rest, version, is_debug):
        self.version_dict = {}
        self.rest = rest
        self.utils = utils.PowerMaxUtils()
        self.volume_trace_list = []
        self.is_debug = is_debug
        self.powermax_driver_version = version

    def _update_platform(self):
        """Update the platform."""
        try:
            self.version_dict['openstack_platform'] = platform.platform()
        except Exception as ex:
            LOG.warning("Unable to determine the platform. "
                        "Exception is %s.", ex)

    def _get_python_version(self):
        """Get the python version."""
        try:
            self.version_dict['python_version'] = platform.python_version()
        except Exception as ex:
            LOG.warning("Unable to determine the python version. "
                        "Exception is %s.", ex)

    def _update_version_from_version_string(self):
        """Update the version from the version string."""
        try:
            self.version_dict['openstack_version'] = (
                version.version_info.version_string())
        except Exception as ex:
            LOG.warning("Unable to determine the OS version. "
                        "Exception is %s.", ex)

    def _update_release_from_release_string(self):
        """Update the release from the release string."""
        try:
            self.version_dict['openstack_release'] = (
                version.version_info.release_string())
        except Exception as ex:
            LOG.warning("Unable to get release info. "
                        "Exception is %s.", ex)

    @staticmethod
    def _get_version_info_version():
        """Gets the version.

        :returns: string -- version
        """
        return version.version_info.version

    def _update_info_from_version_info(self):
        """Update class variables from version info."""
        try:
            ver = self._get_version_info_version()
            if ver:
                self.version_dict['openstack_version'] = ver
        except Exception as ex:
            LOG.warning("Unable to get version info. "
                        "Exception is %s.", ex)

    def _update_openstack_info(self):
        """Update openstack info."""
        self._update_version_from_version_string()
        self._update_release_from_release_string()
        self._update_platform()
        self._get_python_version()
        # Some distributions override with more meaningful information
        self._update_info_from_version_info()

    def _update_array_info(self, serial_number):
        """Update PowerMax/VMAX info.

        :param serial_number: the serial number of the array
        """
        u4p_version_dict = (
            self.rest.get_unisphere_version())
        self.version_dict['unisphere_for_powermax_version'] = (
            u4p_version_dict['version'])
        self.version_dict['serial_number'] = serial_number
        array_info_dict = self.rest.get_array_detail(serial_number)
        self.version_dict['storage_firmware_version'] = (
            array_info_dict['ucode'])
        self.version_dict['storage_model'] = array_info_dict['model']
        self.version_dict['powermax_cinder_driver_version'] = (
            self.powermax_driver_version)

    @debug_required
    def gather_version_info(self, serial_number):
        """Gather info on the array

        :param serial_number: the serial number of the array
        :returns: version_dict
        """
        try:
            self._update_openstack_info()
            self._update_array_info(serial_number)
            self.print_pretty_table(self.version_dict)
        except Exception as ex:
            LOG.warning("Unable to gather version info. "
                        "Exception is %s.", ex)
        return self.version_dict

    @debug_required
    def gather_volume_info(
            self, volume_id, successful_operation, append, **kwargs):
        """Gather volume information.

        :param volume_id: the unique volume id key
        :param successful_operation: the operation e.g "create"
        :param append: append flag
        :param kwargs: variable length argument list
        :returns: datadict
        """
        volume_trace_dict = {}
        volume_key_value = {}
        datadict = {}
        try:
            volume_trace_dict = self._fill_volume_trace_dict(
                volume_id, successful_operation, append, **kwargs)
            volume_trace_dict['volume_updated_time'] = (
                datetime.datetime.fromtimestamp(
                    int(time.time())).strftime('%Y-%m-%d %H:%M:%S'))
            volume_key_value[volume_id] = volume_trace_dict
            if not self.volume_trace_list:
                self.volume_trace_list.append(volume_key_value.copy())
            else:
                self._consolidate_volume_trace_list(
                    volume_id, volume_trace_dict, volume_key_value)
            for datadict in list(self.volume_trace_list):
                if volume_id in datadict:
                    if not append:
                        self.volume_trace_list.remove(datadict)
                    return datadict
        except Exception as ex:
            LOG.warning("Exception in gather volume metadata. "
                        "Exception is %s.", ex)
        return datadict

    def _fill_volume_trace_dict(
            self, volume_id, successful_operation, append, **kwargs):
        """Populates a dictionary with key value pairs

        :param volume_id: the unique volume id key
        :param successful_operation: the operation e.g "create"
        :param append: append flag
        :param kwargs: variable length argument list
        :returns: my_volume_trace_dict
        """
        param_dict = locals()
        my_volume_trace_dict = {}
        for k, v in param_dict.items():
            if self._param_condition(k, v):
                my_volume_trace_dict[k] = v
            if k == 'kwargs':
                for k2, v2 in v.items():
                    if self._param_condition(k2, v2):
                        my_volume_trace_dict[k2] = v2
                    elif k2 == 'mv_list' and v2:
                        for i, item in enumerate(v2, 1):
                            my_volume_trace_dict["masking_view_%d" % i] = item
                    elif k2 == 'sg_list' and v2:
                        for i, item in enumerate(v2, 1):
                            my_volume_trace_dict["storage_group_%d" % i] = item
        return my_volume_trace_dict

    def _param_condition(self, key, value):
        """Determines condition for inclusion.

        :param key: the key
        :param value: the value

        :returns: True or False
        """
        exclude_list = ('self', 'append', 'mv_list', 'sg_list')
        return (value is not None and key not in exclude_list and
                not isinstance(value, (dict,
                                       types.FunctionType,
                                       type)))

    @debug_required
    def print_pretty_table(self, datadict):
        """Prints the data in the dict.

        :param datadict: the data dictionary
        """
        t = prettytable.PrettyTable(['Key', 'Value'])
        for k, v in datadict.items():
            if v is not None:
                t.add_row([k, v])

        LOG.debug('\n%(output)s\n', {'output': t})

    def _consolidate_volume_trace_list(
            self, volume_id, volume_trace_dict, volume_key_value):
        """Consolidate data into self.volume_trace_list

        :param volume_id: the unique volume identifier
        :param volume_trace_dict: the existing dict
        :param volume_key_value: the volume id key and dict value
        """
        is_merged = False
        for datadict in list(self.volume_trace_list):
            if volume_id in datadict:
                for key, dict_value in datadict.items():
                    merged_dict = (
                        self.utils.merge_dicts(
                            volume_trace_dict, dict_value))
                    self.volume_trace_list.remove(datadict)
                    volume_key_value[volume_id] = merged_dict
                    self.volume_trace_list.append(volume_key_value.copy())
                    is_merged = True
        if not is_merged:
            self.volume_trace_list.append(volume_key_value.copy())

    @debug_required
    def update_volume_info_metadata(self, datadict, version_dict):
        """Get update volume metadata with volume info

        :param datadict: volume info key value pairs
        :param version_dict: version dictionary
        :returns: volume_metadata
        """
        return self.utils.merge_dicts(
            version_dict, *datadict.values())

    @debug_required
    def capture_attach_info(
            self, volume, extra_specs, masking_view_dict, host,
            is_multipath, is_multiattach):
        """Captures attach info in volume metadata

        :param volume: the volume object
        :param extra_specs: extra specifications
        :param masking_view_dict: masking view dict
        :param host: host
        :param is_multipath: is mulitipath flag
        :param is_multiattach: is multi attach
        """
        mv_list, sg_list = [], []
        child_storage_group, parent_storage_group = None, None
        initiator_group, port_group = None, None

        if is_multiattach:
            successful_operation = 'multi_attach'
            mv_list = masking_view_dict['mv_list']
            sg_list = masking_view_dict['sg_list']
        else:
            successful_operation = 'attach'
            child_storage_group = masking_view_dict[utils.SG_NAME]
            parent_storage_group = masking_view_dict[utils.PARENT_SG_NAME]
            initiator_group = masking_view_dict[utils.IG_NAME]
            port_group = masking_view_dict[utils.PORTGROUPNAME]

        sl, wl = self.utils.get_service_level_workload(extra_specs)
        datadict = self.gather_volume_info(
            volume.id, successful_operation, False,
            serial_number=extra_specs[utils.ARRAY],
            service_level=sl, workload=wl, srp=extra_specs[utils.SRP],
            masking_view=masking_view_dict[utils.MV_NAME],
            child_storage_group=child_storage_group,
            parent_storage_group=parent_storage_group,
            initiator_group=initiator_group,
            port_group=port_group,
            host=host, is_multipath=is_multipath,
            identifier_name=self.utils.get_volume_element_name(volume.id),
            openstack_name=volume.display_name,
            mv_list=mv_list, sg_list=sg_list)

        volume_metadata = self.update_volume_info_metadata(
            datadict, self.version_dict)

        self.print_pretty_table(volume_metadata)

    @debug_required
    def capture_detach_info(
            self, volume, extra_specs, device_id, mv_list, sg_list):
        """Captures detach info in volume metadata

        :param volume: the volume object
        :param extra_specs: extra specifications
        :param device_id: masking view dict
        :param mv_list: masking view list
        :param sg_list: storage group list
        """
        default_sg = self.utils.derive_default_sg_from_extra_specs(extra_specs)
        sl, wl = self.utils.get_service_level_workload(extra_specs)
        datadict = self.gather_volume_info(
            volume.id, 'detach', False, device_id=device_id,
            serial_number=extra_specs[utils.ARRAY],
            service_level=sl, workload=wl,
            srp=extra_specs[utils.SRP], default_sg_name=default_sg,
            identifier_name=self.utils.get_volume_element_name(volume.id),
            openstack_name=volume.display_name,
            mv_list=mv_list, sg_list=sg_list
        )
        volume_metadata = self.update_volume_info_metadata(
            datadict, self.version_dict)
        self.print_pretty_table(volume_metadata)

    @debug_required
    def capture_extend_info(
            self, volume, new_size, device_id, extra_specs, array):
        """Capture extend info in volume metadata

        :param volume: the volume object
        :param new_size: new size
        :param device_id: device id
        :param extra_specs: extra specifications
        :param array: array serial number
        """
        default_sg = self.utils.derive_default_sg_from_extra_specs(extra_specs)
        sl, wl = self.utils.get_service_level_workload(extra_specs)
        datadict = self.gather_volume_info(
            volume.id, 'extend', False, volume_size=new_size,
            device_id=device_id,
            default_sg_name=default_sg, serial_number=array,
            service_level=sl, workload=wl,
            srp=extra_specs[utils.SRP],
            identifier_name=self.utils.get_volume_element_name(volume.id),
            openstack_name=volume.display_name,
            is_compression_disabled=self.utils.is_compression_disabled(
                extra_specs))
        volume_metadata = self.update_volume_info_metadata(
            datadict, self.version_dict)
        self.print_pretty_table(volume_metadata)

    @debug_required
    def capture_snapshot_info(
            self, source, extra_specs, successful_operation, last_ss_name):
        """Captures snapshot info in volume metadata

        :param source: the source volume object
        :param extra_specs: extra specifications
        :param successful_operation: snapshot operation
        :param last_ss_name: the last snapshot name
        """
        if isinstance(source, volume.Volume):
            if 'create' in successful_operation:
                snapshot_count = six.text_type(len(source.snapshots))
            else:
                snapshot_count = six.text_type(len(source.snapshots) - 1)
            default_sg = (
                self.utils.derive_default_sg_from_extra_specs(extra_specs))
            sl, wl = self.utils.get_service_level_workload(extra_specs)
            datadict = self.gather_volume_info(
                source.id, successful_operation, False,
                volume_size=source.size,
                default_sg_name=default_sg,
                serial_number=extra_specs[utils.ARRAY],
                service_level=sl, workload=wl,
                srp=extra_specs[utils.SRP],
                identifier_name=(
                    self.utils.get_volume_element_name(source.id)),
                openstack_name=source.display_name,
                snapshot_count=snapshot_count,
                last_ss_name=last_ss_name)
            volume_metadata = self.update_volume_info_metadata(
                datadict, self.version_dict)
            self.print_pretty_table(volume_metadata)

    @debug_required
    def capture_modify_group(
            self, group_name, group_id, add_vols, remove_volumes, array):
        """Captures group info after a modify operation

        :param group_name: group name
        :param group_id: group id
        :param add_vols: add volume list
        :param remove_volumes: remove volume list
        :param array: array serial number
        """
        if not self.version_dict:
            self.version_dict = self.gather_version_info(array)
        for add_vol in add_vols:
            datadict = self.gather_volume_info(
                add_vol.id, 'addToGroup', True,
                group_name=group_name, group_id=group_id)
            add_volume_metadata = self.update_volume_info_metadata(
                datadict, self.version_dict)
            self.print_pretty_table(add_volume_metadata)

        for remove_volume in remove_volumes:
            datadict = self.gather_volume_info(
                remove_volume.id, 'removeFromGroup', True,
                group_name='Removed from %s' % group_name,
                group_id='Removed from %s' % group_id)
            remove_volume_metadata = self.update_volume_info_metadata(
                datadict, self.version_dict)
            self.print_pretty_table(remove_volume_metadata)

    @debug_required
    def capture_create_volume(
            self, device_id, volume, group_name, group_id, extra_specs,
            rep_info_dict, successful_operation, source_snapshot_id=None,
            source_device_id=None, temporary_snapvx=None):
        """Captures create volume info in volume metadata

        :param device_id: device id
        :param volume: volume object
        :param group_name: group name
        :param group_id: group id
        :param extra_specs: additional info
        :param rep_info_dict: information gathered from replication
        :param successful_operation: the type of create operation
        :param source_snapshot_id: the source snapshot id

        :returns: volume_metadata dict
        """
        rdf_group_no, target_name, remote_array, target_device_id = (
            None, None, None, None)
        rep_mode, replication_status, rdf_group_label, use_bias = (
            None, None, None, None)
        target_array_model = None
        if rep_info_dict:
            rdf_group_no = rep_info_dict['rdf_group_no']
            target_name = rep_info_dict['target_name']
            remote_array = rep_info_dict['remote_array']
            target_device_id = rep_info_dict['target_device_id']
            rep_mode = rep_info_dict['rep_mode']
            replication_status = rep_info_dict['replication_status']
            rdf_group_label = rep_info_dict['rdf_group_label']
            if utils.METROBIAS in extra_specs:
                use_bias = extra_specs[utils.METROBIAS]
            target_array_model = rep_info_dict['target_array_model']

        default_sg = self.utils.derive_default_sg_from_extra_specs(
            extra_specs, rep_mode)
        sl, wl = self.utils.get_service_level_workload(extra_specs)
        datadict = self.gather_volume_info(
            volume.id, successful_operation, True, volume_size=volume.size,
            device_id=device_id,
            default_sg_name=default_sg,
            serial_number=extra_specs[utils.ARRAY],
            service_level=sl, workload=wl,
            srp=extra_specs[utils.SRP],
            identifier_name=self.utils.get_volume_element_name(volume.id),
            openstack_name=volume.display_name,
            source_volid=volume.source_volid,
            group_name=group_name, group_id=group_id,
            rdf_group_no=rdf_group_no,
            target_name=target_name, remote_array=remote_array,
            target_device_id=target_device_id,
            source_snapshot_id=source_snapshot_id,
            rep_mode=rep_mode, replication_status=replication_status,
            rdf_group_label=rdf_group_label, use_bias=use_bias,
            is_compression_disabled=self.utils.is_compression_disabled(
                extra_specs),
            source_device_id=source_device_id,
            temporary_snapvx=temporary_snapvx,
            target_array_model=target_array_model)
        volume_metadata = self.update_volume_info_metadata(
            datadict, self.version_dict)
        self.print_pretty_table(volume_metadata)

    @debug_required
    def gather_replication_info(
            self, volume_id, successful_operation, append, **kwargs):
        """Gathers replication information

        :param volume_id: volume id
        :param successful_operation: the successful operation type
        :param append: boolean
        :param **kwargs: variable length of arguments
        :returns: rep_dict
        """
        return self._fill_volume_trace_dict(
            volume_id, successful_operation, append, **kwargs)

    @debug_required
    def capture_failover_volume(
            self, volume, target_device, remote_array, rdf_group, device_id,
            array, extra_specs, failover, vol_grp_name,
            replication_status, rep_mode):
        """Captures failover info in volume metadata

        :param volume: volume object
        :param target_device: the device to failover to
        :param remote_array: the array to failover to
        :param rdf_group: the rdf group
        :param device_id: the device to failover from
        :param array: the array to failover from
        :param extra_specs: additional info
        :param failover: failover flag
        :param vol_grp_name: async group name
        :param replication_status: volume replication status
        :param rep_mode: replication mode
        """
        operation = "Failover" if failover else "Failback"
        sl, wl = self.utils.get_service_level_workload(extra_specs)
        datadict = self.gather_volume_info(
            volume.id, operation, True, volume_size=volume.size,
            device_id=target_device,
            serial_number=remote_array,
            service_level=sl, workload=wl,
            srp=extra_specs[utils.SRP],
            identifier_name=self.utils.get_volume_element_name(volume.id),
            openstack_name=volume.display_name,
            source_volid=volume.source_volid,
            rdf_group_no=rdf_group, remote_array=array,
            target_device_id=device_id, vol_grp_name=vol_grp_name,
            replication_status=replication_status, rep_mode=rep_mode
        )

        self.version_dict = (
            self.gather_version_info(remote_array))
        volume_metadata = self.update_volume_info_metadata(
            datadict, self.version_dict)
        self.print_pretty_table(volume_metadata)

    @debug_required
    def capture_manage_existing(
            self, volume, rep_info_dict, device_id, extra_specs):
        """Captures manage existing info in volume metadata

        :param volume: volume object
        :param rep_info_dict: information gathered from replication
        :param device_id: the PowerMax/VMAX device id
        :param extra_specs: the extra specs
        """
        successful_operation = "manage_existing_volume"
        rdf_group_no, target_name, remote_array, target_device_id = (
            None, None, None, None)
        rep_mode, replication_status, rdf_group_label = (
            None, None, None)
        if rep_info_dict:
            rdf_group_no = rep_info_dict['rdf_group_no']
            target_name = rep_info_dict['target_name']
            remote_array = rep_info_dict['remote_array']
            target_device_id = rep_info_dict['target_device_id']
            rep_mode = rep_info_dict['rep_mode']
            replication_status = rep_info_dict['replication_status']
            rdf_group_label = rep_info_dict['rdf_group_label']

        default_sg = self.utils.derive_default_sg_from_extra_specs(
            extra_specs, rep_mode)
        sl, wl = self.utils.get_service_level_workload(extra_specs)
        datadict = self.gather_volume_info(
            volume.id, successful_operation, True, volume_size=volume.size,
            device_id=device_id,
            default_sg_name=default_sg,
            serial_number=extra_specs[utils.ARRAY],
            service_level=sl, workload=wl,
            srp=extra_specs[utils.SRP],
            identifier_name=self.utils.get_volume_element_name(volume.id),
            openstack_name=volume.display_name,
            source_volid=volume.source_volid,
            rdf_group_no=rdf_group_no,
            target_name=target_name, remote_array=remote_array,
            target_device_id=target_device_id,
            rep_mode=rep_mode, replication_status=replication_status,
            rdf_group_label=rdf_group_label
        )
        volume_metadata = self.update_volume_info_metadata(
            datadict, self.version_dict)
        self.print_pretty_table(volume_metadata)

    @debug_required
    def capture_retype_info(
            self, volume, device_id, array, srp, target_slo,
            target_workload, target_sg_name, is_rep_enabled, rep_mode,
            is_compression_disabled):
        """Captures manage existing info in volume metadata

        :param volume_id: volume identifier
        :param volume_size: volume size
        :param device_id: the PowerMax/VMAX device id
        :param array: the PowerMax/VMAX serialnumber
        :param srp: PowerMax/VMAX SRP
        :param target_slo: volume name
        :param target_workload: the PowerMax/VMAX device id
        :param is_rep_enabled: replication enabled flag
        :param rep_mode: replication mode
        :param is_compression_disabled: compression disabled flag
        """
        successful_operation = "retype"
        if not target_slo:
            target_slo, target_workload = 'None', 'None'
        datadict = self.gather_volume_info(
            volume.id, successful_operation, False, volume_size=volume.size,
            device_id=device_id,
            target_sg_name=target_sg_name,
            serial_number=array,
            target_service_level=target_slo,
            target_workload=target_workload,
            srp=srp,
            identifier_name=self.utils.get_volume_element_name(volume.id),
            openstack_name=volume.display_name,
            is_rep_enabled=('yes' if is_rep_enabled else 'no'),
            rep_mode=rep_mode, is_compression_disabled=(
                True if is_compression_disabled else False))
        if not is_rep_enabled:
            delete_list = ['rdf_group_no', 'rep_mode', 'target_array_model',
                           'service_level', 'remote_array', 'target_device_id',
                           'replication_status', 'rdf_group_label']
            self.utils.delete_values_from_dict(datadict, delete_list)
            update_list = [('default_sg_name', 'source_sg_name'),
                           ('service_level', 'source_service_level')]
            self.utils.update_values_in_dict(datadict, update_list)

        volume_metadata = self.update_volume_info_metadata(
            datadict, self.version_dict)
        self.print_pretty_table(volume_metadata)

    @debug_required
    def capture_delete_info(self, volume):
        """Captures delete info in volume metadata

        :param volume: the volume object
        """
        datadict = self.gather_volume_info(
            volume.id, 'delete', False,
            identifier_name=self.utils.get_volume_element_name(volume.id),
            openstack_name=volume.display_name)
        volume_metadata = self.update_volume_info_metadata(
            datadict, self.version_dict)
        self.print_pretty_table(volume_metadata)



# Copyright (c) 2017 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from cinder.policies import base

CREATE_POLICY = 'group:create'
UPDATE_POLICY = 'group:update'
GET_POLICY = 'group:get'
GET_ALL_POLICY = 'group:get_all'
GROUP_ATTRIBUTES_POLICY = 'group:group_project_attribute'


groups_policies = [
    policy.DocumentedRuleDefault(
        name=GET_ALL_POLICY,
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="List groups.",
        operations=[
            {
                'method': 'GET',
                'path': '/groups'
            },
            {
                'method': 'GET',
                'path': '/groups/detail'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=CREATE_POLICY,
        check_str="",
        description="Create group.",
        operations=[
            {
                'method': 'POST',
                'path': '/groups'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=GET_POLICY,
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Show group.",
        operations=[
            {
                'method': 'GET',
                'path': '/groups/{group_id}'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=UPDATE_POLICY,
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Update group.",
        operations=[
            {
                'method': 'PUT',
                'path': '/groups/{group_id}'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=GROUP_ATTRIBUTES_POLICY,
        check_str=base.RULE_ADMIN_API,
        description="List groups or show group with project attributes.",
        operations=[
            {
                'method': 'GET',
                'path': '/groups/{group_id}'
            },
            {
                'method': 'GET',
                'path': '/groups/detail'
            }
        ]),
]


def list_rules():
    return groups_policies
